# coding: utf-8

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import requests
import json
import os
import pickle

from NORDic.UTILS.LINCS_utils import *
from NORDic.UTILS.utils_state import binarize_experiments

def profiles2signatures(profiles_df, user_key, path_to_lincs, save_fname, backgroundfile=False, selection="distil_ss", thres=0.5, bin_method="binary", nbackground_limits=(4,30), quiet=False):
    '''
    Convert experimental profiles into signatures (1 for control samples, 1 for treated ones)

    ...

    Parameters
    ----------
    profiles_df : Pandas DataFrame
        rows/[genes+annotations] x columns/[samples]
    user_key : Python character string
        LINCS L1000 user API key
    path_to_lincs : Python character string
        path to local LINCS L1000 files
    save_fname : Python character string
        path to save normalized expression profiles per cell line
    background_file : Python bool
        [default=False] : retrieves from LINCS L1000 supplementary expression values if set to True to compute more precise basal gene expression levels
    selection : Python character string
        [default="distil_ss"] : LINCS L1000 metric to maximize for the "background" data
    thres : Python float
        [default=0.5] : threshold for cutoff normalized gene expression values (in [0,0.5])
    bin_method : Python character string
        [default="binary"] : binarization approach
    nbackground_limits : Python integer tuple
        [default=(4,30)] : lower and upper bounds on the number of profiles for the background expression data
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    signatures_df\tPandas DataFrame
        rows/[genes] x columns/[signature ID]
    '''
    assert thres >= 0 and thres <= 0.5
    assert path_to_lincs
    assert user_key
    assert bin_method in ["binary", "binary_CD"]
    selection_min, selection_max = nbackground_limits
    assert selection_min > 0
    assert selection_max >= selection_min
    cell_lines = list(set([x for x in list(profiles_df.loc["cell_line"])]))
    signatures_list, conditions_spec = [], ["perturbed", "perturbation"]
    add_rows_profiles = ["annotation", "perturbed", "perturbation", "cell_line", "sigid", "interference_scale"]
    for icell, cell in enumerate(cell_lines):
        cell_save_fname = save_fname+"_"+cell+"_selection="+selection+".csv"
        ## 1. For each cell line, split control and treated samples into two Pandas DataFrames
        profiles__df = profiles_df[[profiles_df.columns[ix] for ix, x in enumerate(profiles_df.loc["cell_line"]) if (x==cell)]]
        initial_cols = profiles__df.loc["annotation"].apply(pd.to_numeric).values==1
        final_profiles_df = profiles__df[profiles__df.columns[~initial_cols]]
        conditions = ["_".join(list(final_profiles_df[idx].loc[conditions_spec])) for idx in final_profiles_df.columns]
        initial_profiles = profiles__df[profiles__df.columns[initial_cols]].loc[[idx for idx in profiles__df.index if (idx not in add_rows_profiles)]].apply(pd.to_numeric)
        initial_profiles.columns = ["Ctrl_rep%d" % (i+1) for i in range(initial_profiles.shape[1])]
        final_profiles_df.columns = [x+"_%d" % (ix+1) for ix, x in enumerate(conditions)]
        final_profiles = final_profiles_df.loc[[idx for idx in final_profiles_df.index if (idx not in add_rows_profiles)]].apply(pd.to_numeric)
        assert final_profiles.shape[1]==np.sum(profiles__df.loc["annotation"].apply(pd.to_numeric).values==2)
        assert initial_profiles.shape[1]==np.sum(profiles__df.loc["annotation"].apply(pd.to_numeric).values==1)
        if (not quiet):
            print("<UTILS_EXP> Cell %s (%d/%d)" % (cell, icell+1, len(cell_lines)), end="... ")
        ## 1'. If required, retrieve background expression data corresponding to the considered cell line
        if (not os.path.exists(cell_save_fname) and backgroundfile):
            endpoint = "sigs"
            method = "filter"
            params = {
                "where": {"cell_id": cell, "pert_type": "trt_sh"}, 
                "fields": ["distil_cc_q75", selection, "pct_self_rank_q25", "distil_id", "brew_prefix"]
            }
            request_url = build_url(endpoint, method, params=params, user_key=user_key)
            response = requests.get(request_url)
            assert response.status_code == 200
            data = json.loads(response.text)
            data = [dt for dt in data if (("distil_id" in dt) and ("distil_cc_q75" in dt) and ("pct_self_rank_q25" in dt))]
            ## Select only "gold" signatures, as defined by LINCS L1000 
            ## https://clue.io/connectopedia/glossary#is_gold
            data = [dt for dt in data if (len(dt["distil_id"])>1)]
            assert len(data)>0
            data_gold = [dt for dt in data if ((dt["distil_cc_q75"]>=0.2) and (dt["pct_self_rank_q25"]<=0.05))]
            if (len(data_gold)>0):
                data = data_gold
            ## Select profiles maximizing the "selection" criterion
            mselection = np.min([dt[selection] for dt in data])
            max_selection = np.argsort(np.array([dt[selection] if (dt[selection]>=selection_min) else mselection for dt in data]))
            max_selection = max_selection[-min(selection_max,len(max_selection)):]
            vals_selection = [dt[selection] for dt in [data[i] for i in max_selection]]
            if (not quiet):
                print("<UTILS_EXP> %d (good) profiles | %d (best) profiles (capped at 50 or min>=%d) (%s max=%.3f, min=%.3f)" % (len(data), len(max_selection), selection_min, selection, np.max(vals_selection), np.min(vals_selection)))
            bkgrnd = create_restricted_drug_signatures([did for dt in [data[i] for i in max_selection] for did in dt["distil_id"]], [int(g) for g in list(profiles__df.index) if (g not in add_rows_profiles)], path_to_lincs, which_lvl=[3], strict=False)
            bkgrnd.index = [int(g) for g in bkgrnd.index]
            bkgrnd.to_csv(cell_save_fname)
        elif (backgroundfile):
            bkgrnd = pd.read_csv(cell_save_fname, index_col=0)
        ## 2. Aggregate replicates by median values for signature ~ initial condition
        initial_profile = initial_profiles.median(axis=1)
        initial_profile = initial_profile.loc[~initial_profile.duplicated()]
        ## 3. Aggregate values across probes of the same gene
        final_profile = final_profiles.T
        final_profile.index = ["_".join(list(final_profiles_df[idx].loc[conditions_spec])) for idx in final_profiles_df.columns]
        data = final_profile.groupby(level=0).median().T
        data = data.loc[~data.index.duplicated()]
        data["initial"] = initial_profile.loc[[i for i in data.index if (i in initial_profile.index)]]
        data.index = data.index.astype(int)
        if (backgroundfile):
            data = data.join(bkgrnd, how="inner")
        ## 4. Binarize profiles
        signatures = binarize_experiments(data, thres=thres, method=bin_method.split("_CD")[0], strict=not ('CD' in bin_method))
        signatures = signatures[list(set(conditions))+["initial"]]
        if ("_CD" in bin_method):
            for c in conditions:
                df = pd.concat((final_profiles[[col for col in final_profiles.columns if (c in col)]], initial_profiles), axis=1)
                samples = [int("Ctrl_" in col)+1 for col in df.columns]
                sigs = binarize_via_CD(df, samples=samples, binarize=1, nperm=10000)
                signatures[c] = list(sigs["aggregated"])
        signatures.columns = [s+"_"+cell for s in signatures.columns]
        signatures_list.append(signatures)
    signatures_df = signatures_list[0].join(signatures_list[1:], how="outer")
    return signatures_df

def get_experimental_constraints(file_folder,cell_lines, pert_types, pert_di, taxon_id, selection, user_key, path_to_lincs, thres_iscale=None, nsigs=2, quiet=False):
    '''
    Retrieve experimental profiles from the provided cell lines, perturbation types, list of genes, in the given species (taxon ID)

    ...

    Parameters
    ----------
    file_folder : Python character string
        folder where to store intermediary results
    cell_lines : Python character string list
        cell lines present in LINCS L1000
    pert_types : Python character string list
        types of perturbations as supported by LINCS L1000
    pert_di : Python dictionary
        (keys=Python character string, values=Python integer) associates HUGO gene symbols to their EntrezGene IDs
    taxon_id : Python integer
        NCBI taxonomy ID
    selection : Python character string
        LINCS L1000 metric to maximize
    user_key : Python character string
        LINCS L1000 user API key
    path_to_lincs : Python character string
        path to local LINCS L1000 files
    thres_iscale : Python float or None
        [default=None] : lower threshold on the interference scale which quantifies the success of a genetic experiment
    nsigs : Python integer
        [default=2] : minimal number of profiles per experiment and condition
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    signatures : Pandas DataFrame
        rows/[genes+annotations] x columns/[profile/signature IDs]
    '''
    assert str(thres_iscale) == "None" or thres_iscale >= 0
    assert len(cell_lines) > 0
    assert len(pert_types) > 0
    assert nsigs > 1
    entrez_ids, pert_inames = list(pert_di.values()), list(pert_di.keys())
    signatures, perturbed_genes = [], []
    ## 1. For each cell line
    for il, line in enumerate(cell_lines):
        ids = {}
        pert_types_ = [p for p in pert_types if (p != "ctl_untrt")]
        for ip, pert_type in enumerate(pert_types_):
            endpoint = "sigs"
            method = "filter"
            ## 2. Build the additional parameters for the LINCS L1000 request
            ## 2a. Select the type of perturbation and cell line
            where = {"pert_type": pert_type, "cell_id": line}
            ## 2b. Return the HUGO gene symbol of the perturbed gene, the perturbagen dose/type/cell line/exposure time
            ## and distil_id (unique to profile), brew_prefix (unique to set of replicates)
            params = { "where": where,
                    "fields": ["pert_iname", "pert_idose", "pert_type", "cell_id", "pert_itime", "distil_id", "brew_prefix"] }
            ## 3. Create URL based on those parameters
            #request_url = build_url(endpoint, method, params=params, user_key=user_key)
            ## 4. Retrieve the set of experiments matching those constraints, with enough replicates (>=nsigs)
            ## /!\ restriction due to limit=1000 in LINCS L1000 requests
            #data_pert_ = post_request(request_url, quiet=True)
            #data_pert_ = [dt for dt in data_pert_ if ((dt["pert_iname"] in pert_inames) and (len(dt["distil_id"])>=nsigs))]
            data_pert_fname = file_folder+"data_pert_cellline=%s_perttype=%s.pck" % (line, pert_type)
            if (not os.path.exists(data_pert_fname)):
                data_pert_, seen_genes = [], []
            else:
                with open(data_pert_fname, "rb") as f:
                    results = pickle.load(f)
                data_pert_, seen_genes = results["data_pert"], results["seen_genes"]
            for ig, gene in enumerate(pert_inames):
                if (gene in seen_genes):
                    continue
                if (not quiet):
                    print("<UTILS_EXP> Gene %d/%d Cell %d/%d Type %d/%d" % (ig+1, len(pert_inames), il+1, len(cell_lines), ip+1, len(pert_types_)))
                p = {'where': {"pert_type": pert_type, "cell_id": line, "pert_iname": gene}, "fields": params["fields"]}
                res_pert_ = post_request(build_url(endpoint, method, params=p, user_key=user_key), quiet=True, pause_time=0.3)
                if (len(res_pert_)>0 and not quiet):
                    print("\t<UTILS_EXP> (%s,%s,%s): %d" % (gene, line, pert_type, len(res_pert_)))
                data_pert_ += [dt for dt in res_pert_ if (len(dt["distil_id"])>=nsigs)]
                seen_genes.append(gene)
                with open(data_pert_fname, "wb") as f:
                    pickle.dump({"data_pert": data_pert_, "seen_genes": seen_genes}, f)
            with open(data_pert_fname, "rb") as f:
                results = pickle.load(f)
            data_pert_ = results["data_pert"]
            sigs_fname = file_folder+"sigs_cellline=%s_perttype=%s.csv" % (line, pert_type)
            if (os.path.exists(sigs_fname)):
                sigs = pd.read_csv(sigs_fname, index_col=0)
                signatures = [sigs[[c for c in sigs.columns if (sigs.loc["perturbed"][c]==gene)]] for gene in list(sigs.T["perturbed"])]
                perturbed_genes = list(set(list(sigs.loc["perturbed"])))
            for data in data_pert_:
                entrez_id = pert_di[data["pert_iname"]]
                if (not quiet):
                    print("<UTILS_EXP> %d experiments so far" % len(signatures))
                treatment, perturbation = str(data["pert_iname"]), "OE" if ("_oe" in pert_type) else "KD"
                ## avoid duplicates
                if (treatment in perturbed_genes and not quiet):
                    print("\t<UTILS_EXP> Duplicated treatment:%s, cell:%s, type:%s" % (treatment, str(data["cell_id"]), str(data["pert_type"])))
                    continue
                elif (not quiet):
                    print("\t<UTILS_EXP> Treatment %s (entrez_id %d)... " % (treatment, entrez_id), end="")
                ## 6. Returns control & treated profiles from LINCS L1000
                sigs = get_treated_control_dataset(treatment, pert_type, line, {}, entrez_ids, taxon_id, user_key, path_to_lincs, entrez_id=entrez_id,
                        which_lvl=[3], nsigs=nsigs, same_plate=True, selection=selection, quiet=quiet, trim_w_interference_scale=True)
                if (sigs is None or len(sigs)==0):
                    continue
                if (not quiet):
                    print("... %d genes, %d profiles" % (sigs.shape[0]-3, sigs.shape[1]))
                perturbed_genes.append(treatment)
                sigs.loc["perturbed"] = [treatment]*sigs.shape[1]
                sigs.loc["perturbation"] = [perturbation]*sigs.shape[1]
                sigs.loc["cell_line"] = [line]*sigs.shape[1]
                sigs.loc["sigid"] = list(sigs.columns)
                nexp = len(signatures)+1
                sigs.columns = ["Exp"+str(nexp)+":"+str(i)+"-rep"+str(ai+1) for ai, i in enumerate(list(sigs.loc["annotation"]))]
                if (len(signatures)==0):
                    sigs.to_csv(sigs_fname)
                else:
                    signatures[0].join(signatures[1:], how="outer").to_csv(sigs_fname)
                signatures.append(sigs)
    if (len(signatures)==0):
        return pd.DataFrame([], index=pert_inames)
    signatures = signatures[0].join(signatures[1:], how="outer")
    signatures = signatures.loc[~signatures.index.duplicated()]
    return signatures



# coding: utf-8

import bonesis
import mpbn
import numpy as np
import pandas as pd
from itertools import product
from functools import reduce
import sys
from zipfile import ZipFile
import os
from glob import glob
from tqdm import tqdm
from subprocess import call as sbcall
from copy import deepcopy
from sklearn.linear_model import LogisticRegression as Logit

from NORDic.UTILS.utils_state import quantile_normalize

def desirability(x, f_weight_di, A=0, B=1):
    '''
    Harrington's desirability function, used by [1]
    Converts a list of functions to maximize into a single scalar function to maximize with values in [@A,@B]

    [1] http://ceur-ws.org/Vol-2488/paper17.pdf
        https://cran.r-project.org/web/packages/desirability/vignettes/desirability.pdf

    ...

    Parameters
    ----------
    x : datapoint
        any input to functions in f_weight_di
    f_weight_di : Python dictionary
        function with arguments as the same type as x, and associated weight
    A : Python float
        [default=0] : lower bound of the function interval
    B : Python float
        [default=1] : upper bound of the function interval

    Returns
    ----------
    des(x) : Python float
        value of the desirability function at point x
    '''
    values = [np.exp(-np.exp(-((B-A)+weight/(B-A)*funct(x)))) for funct, weight in f_weight_di.items()]
    return float(np.exp(np.mean(np.log(values))))

def DS(influences):
    '''
    Computes the number of edges over the maximum number of possible connections between the nodes in the network

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes]

    Returns
    ----------
    DS : Python float
        network density
    '''
    n = influences.shape[0]
    return 2*np.sum(np.sum(np.triu(np.abs(influences.values)), axis=1), axis=0)/float(n*(n-1))

def CL(influences):
    '''
    Computes the average of node-wise clustering coefficients. The clustering coefficient of a node is the ratio of the degree of the considered node and the maximum possible number of connections such that this node and its current neighbors form a clique

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes]

    Returns
    ----------
    CL : Python float
        network clustering coefficient
    '''
    n = influences.shape[0]
    network_outdegree = np.sum(np.abs(influences.values), axis=1)
    connect = np.sum(np.triu((np.abs(influences)+np.abs(influences).T).values),axis=1)
    max_connect = 0.5*np.multiply(network_outdegree,network_outdegree-1)
    max_connect[max_connect<=0] = 0.5
    numCL = np.sum(np.divide(connect, max_connect),axis=0)
    return numCL/n

def Centr(influences):
    '''
    Computes the network centralization, which is correlated with the similarity of the network to a graph with a star topology

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes]

    Returns
    ----------
    Centr : Python float
        network centralization
    '''
    n = influences.shape[0]
    network_outdegree = np.sum(np.abs(influences.values),axis=1)
    k_max = np.max(network_outdegree)
    return n/(n-2)*(k_max/(n-1)-DS(influences))

def GT(influences):
    '''
    Computes the network heterogeneity, which quantifies the non-uniformity of the node degrees across the network

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes]

    Returns
    ----------
    GT : Python float
        network heterogeneity
    '''
    network_outdegree = np.sum(np.abs(influences.values), axis=1)
    k_mean = np.mean(network_outdegree)
    return np.sqrt(np.var(network_outdegree))/max(k_mean,1)

def general_topological_parameter(influences, weights):
    '''
    Computes the general topological parameter (GTP) associated with the input network

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes]
    weights : Python dictionary of (Python character string x Python float)
        all keys must be in ["DS","CL","Centr","GT"]

    Returns
    ----------
    score : Python float
        score using the Harrington's desirability function
    '''
    assert all([w in ["DS","CL","Centr","GT"] for w in weights])
    missing_index = [x for x in influences.index if (x not in influences.columns)]
    missing_cols = [x for x in influences.columns if (x not in influences.index)]
    if (len(missing_index)>0):
        influences = pd.concat((influences, pd.DataFrame(0, index=missing_index, columns=influences.columns)), axis=0)
    if (len(missing_cols)>0):
        influences = pd.concat((influences, pd.DataFrame(0, index=list(influences.index)+missing_index, columns=missing_cols)), axis=1)
    return desirability(influences, {DS: weights.get("DS",0), CL: weights.get("CL",0), Centr: weights.get("Centr",0), GT: weights.get("GT", 0)})

def get_weakly_connected(network_df, gene_list, index_col="preferredName_A", column_col="preferredName_B", score_col="sscore"):
    '''
    Depth-first search (DFS) on undirected network

    ...

    Parameters
    ----------
    network_df : Pandas DataFrame
        rows/[index] x columns/[["Input","Output"]]
    gene_list : Python character string list
        list of genes (needed to take into account isolated genes in the network)
    index_col : Python character string
        [default="preferredName_A"] : column in network_df (input gene)
    column_col : Python character string
        [default="preferredName_B"] : column in network_df (output gene)
    score_col : Python character string
        [default="sscore"] : column in network_df (edge weight)

    Returns
    ----------
    components : Type of @network_df.loc[network_df.index[0]]["Input"] Python list of Python list
        list of weakly connected components in the network, ordered by decreasing size
    '''
    adjacency = network_df.pivot_table(index=index_col, columns=column_col, values=score_col, aggfunc="mean")
    ## Undirected adjacency matrix
    adjacency[~np.isnan(adjacency)] = 1
    adjacency = adjacency.fillna(0)
    missing_index = [g for g in gene_list if (g not in adjacency.index)]
    if (len(missing_index)>0):
        missing_index_df = pd.DataFrame(0, index=missing_index, columns=adjacency.columns)
        adjacency = pd.concat((adjacency, missing_index_df), axis=0)
    missing_columns = [g for g in gene_list if (g not in adjacency.columns)]
    if (len(missing_columns)>0):
        missing_columns_df = pd.DataFrame(0, index=adjacency.index, columns=missing_columns)
        adjacency = pd.concat((adjacency, missing_columns_df), axis=1)
    assert adjacency.shape[0]==adjacency.shape[1]==len(gene_list)
    N = len(gene_list)
    components = []
    to_visit = [0]
    visited = [False]*N
    while (not all(visited)):
        component = []
        while (True):
            node = to_visit.pop()
            visited[node] = True
            if (node not in component):
                component.append(node)
            children = [list(adjacency.index).index(list(adjacency.columns)[i]) for i in np.argwhere(adjacency.values[node, :] != 0).flatten().tolist()]
            to_visit = [child for child in children if (not visited[child])]+to_visit
            if (len(to_visit)==0):
                break
        components.append(component)
        to_visit = [np.argmin(visited)]
    components = list(sorted(components, key=lambda c : len(c), reverse=True))
    components = [[list(adjacency.index)[n] for n in c] for c in components]
    return components

def get_genes_interactions_from_PPI(ppi, connected=False, score=0, filtering=True, quiet=False):
    '''
    Filtering edges to decrease computational cost while preserving network connectivity (if needed)

    ...

    Parameters
    ----------
    ppi : Pandas DataFrame
        rows/[index] x columns[{"preferredName_A", "preferredName_B", "sign", "directed", "score"]]; sign in {-1,1,2}, directed in {0,1}, score in [0,1]
    connected : Python bool
        [default=True] : if set to True, preserve/enforce connectivity on the final network
    score : Python float
        [default=0] : Lower bound on the edge-associated score
    filtering : Python bool
        [default=True] : Whether to filter out edges by a correlation threshold
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    ppi_accepted : Pandas DataFrame
        rows/[index] x columns/[["Input", "Output"]]
    '''
    assert all([x in [-1,1,2] for x in list(ppi["sign"])])
    assert all([x in [0,1] for x in list(ppi["directed"])])
    assert all([x <= 1 and x >= 0 for x in list(ppi["score"])])
    cols = ['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score']
    assert all([col in cols for col in ppi.columns])
    assert ppi.shape[1]==len(cols)
    ppi = ppi[['preferredName_B', 'preferredName_A', 'sign', 'directed', 'score']]
    ## Remove duplicate edges (preserve the duplicates with highest score)
    ppi.index = ["--".join(x) for x in zip(ppi["preferredName_A"], ppi["preferredName_B"])]
    ppi = ppi.sort_values(by="score", ascending=False)
    ppi = ppi.loc[~ppi.index.duplicated(keep="first")]
    ppi = ppi[["preferredName_A","preferredName_B", 'sign', 'directed', 'score']]
    ppi.index = range(ppi.shape[0])
    Ntotal_edges = ppi.shape[0]
    ## 1. Double edges depending on whether they are marked as "directed"
    undirected_edges = ppi.loc[ppi["directed"]==0]
    Nundirected_edges = undirected_edges.shape[0]
    undirected_edges.columns = ['preferredName_B', 'preferredName_A', 'sign', 'directed', 'score']
    undirected_edges = undirected_edges[['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score']]
    assert all([c==['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score'][i] for i, c in enumerate(undirected_edges.columns)])
    assert undirected_edges.loc[undirected_edges.index[0]]['preferredName_B']==ppi.loc[ppi["directed"]==0].loc[ppi.loc[ppi["directed"]==0].index[0]]['preferredName_A']
    ppi = pd.concat((ppi, undirected_edges),axis=0)[["preferredName_A","preferredName_B","sign","score"]]
    ppi.index = range(ppi.shape[0])
    assert ppi.shape[0]==Ntotal_edges+Nundirected_edges
    ## 2. Double edges depending on whether they have a specific sign
    nonmonotonic_edges = ppi.loc[ppi["sign"]==2]
    Nnonmonotonic_edges = nonmonotonic_edges.shape[0]
    ppi_sign_sum = sum([abs(x) for x in list(ppi["sign"])])
    nonmonotonic_edges = pd.concat((nonmonotonic_edges[[c for c in nonmonotonic_edges.columns if (c!="sign")]], pd.DataFrame([[-1]]*nonmonotonic_edges.shape[0], index=nonmonotonic_edges.index, columns=["sign"])), axis=1)
    nonmonotonic_edges = nonmonotonic_edges[["preferredName_A","preferredName_B","sign","score"]]
    ppi = pd.concat((ppi, nonmonotonic_edges), axis=0)
    ppi["sign"] = [1 if (s==2) else s for s in list(ppi["sign"])]
    ppi.index = range(ppi.shape[0])
    assert ppi.shape[0]==Ntotal_edges+Nundirected_edges+Nnonmonotonic_edges
    assert ppi["sign"].abs().sum().sum()==ppi_sign_sum
    ## 3. Concatenate columns "score" and "sign" as score in [0,1]
    ppi["sscore"] = np.multiply(ppi["score"], ppi["sign"])
    ppi = ppi[["preferredName_A","preferredName_B","sscore"]]
    if (connected):
        components = get_weakly_connected(ppi, list(set(list(ppi["preferredName_A"])+list(ppi["preferredName_B"]))))
        main_component = components[0]
        ## 5. Remove genes which are isolated in the full PPI
        isolated_genes = [g for c in components[1:] for g in c]
        if (len(isolated_genes)>0):
            test_isolated = np.vectorize(lambda x : x in isolated_genes)
            ppi = ppi.loc[~test_isolated(ppi["preferredName_A"])&~test_isolated(ppi["preferredName_B"])]
        t = min(score, ppi["sscore"].abs().max()) if (filtering) else ppi["sscore"].abs().min()
        ppi_accepted = ppi.loc[ppi["sscore"].abs()>=t]
        components = get_weakly_connected(ppi_accepted, main_component)
        isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
        assert len(isolated_genes)+len(components[0])==len(main_component)
        ## 6. Add edges to/from isolated genes in decreasing (absolute) score order until the PPI is connected and all genes in the connected PPI are present
        if (len(isolated_genes)>0):
            ppi_rejected = ppi.loc[ppi["sscore"].abs()<t]
            test_isolated = np.vectorize(lambda x : x in isolated_genes)
            while ((len(isolated_genes)>0) and (ppi_rejected.shape[0]>0)):
                ppi_isolated = ppi_rejected.loc[(test_isolated(ppi_rejected["preferredName_A"])&~test_isolated(ppi_rejected["preferredName_B"]))|(test_isolated(ppi_rejected["preferredName_B"])&~test_isolated(ppi_rejected["preferredName_A"]))]
                t = np.max(ppi_isolated["sscore"].abs())
                add_ppi_isolated = ppi_isolated.loc[ppi_isolated["sscore"].abs()>=t]
                ppi_accepted = pd.concat((ppi_accepted, add_ppi_isolated), axis=0)
                components = get_weakly_connected(ppi_accepted, main_component)
                if (not quiet):
                    print("adding %d edges at t=%f (%d edges in total)" % (add_ppi_isolated.shape[0], t, ppi_accepted.shape[0]))
                isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
                if (not quiet):
                    print("Components %s... (%d isolated genes)" % (str([len(c) for c in components[:3]]),len(isolated_genes)))
                assert len(isolated_genes)+len(components[0])==len(main_component)
                test_isolated = np.vectorize(lambda x : x in isolated_genes)
                ppi_rejected = ppi_rejected.loc[(ppi_rejected["sscore"].abs()<t)|((test_isolated(ppi_rejected["preferredName_A"])|(test_isolated(ppi_rejected["preferredName_B"]))))]
        assert len(components)==1 and len(components[0])==len(main_component)
    else:
        t = min(score, ppi["sscore"].abs().max()) if (filtering) else ppi["sscore"].abs().min()
        ppi_accepted = ppi.loc[ppi["sscore"].abs()>=t]
    ppi_accepted.columns = ["Input","Output","SSign"]
    return ppi_accepted

def build_influences(network_df, tau, beta=1, cor_method="pearson", expr_df=None, accept_nonRNA=False, quiet=False):
    '''
    Filters out (and signs of unsigned) edges based on gene expression 

    ...

    Parameters
    ----------
    network_df : Pandas DataFrame
        rows/[index] x columns/[["Input", "Output", "SSign"]] interactions
    tau : Python float
        threshold on genepairwise expression correlation
    beta : Python integer
        [default=1] : power applied to the adjacency matrix
    cor_method : Python character string
        [default="pearson"] : type of correlation
    expr_df : Pandas DataFrame
        [default=None] : rows/[genes] x columns/[samples] gene expression data
    accept_nonRNA : Python bool
        [default=False] : if set to False, ignores gene names which are not present in expr_df
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes] signed adjacency matrix with only interactions s.t. corr^beta>=tau
    '''
    assert network_df.shape[1]==3 and all([c in ["Input","Output", "SSign"] for c in network_df.columns])
    network = network_df.pivot_table(index="Input", columns="Output", values="SSign", aggfunc="mean")
    missing_index = [g for g in network.columns if (g not in network.index)]
    if (len(missing_index)>0):
        missing_index_df = pd.DataFrame([], index=missing_index, columns=network.columns)
        network = pd.concat((network, missing_index_df), axis=0)
    missing_columns = [g for g in network.index if (g not in network.columns)]
    if (len(missing_columns)>0):
        missing_columns_df = pd.DataFrame([], index=network.index, columns=missing_columns)
        network = pd.concat((network, missing_columns_df), axis=1)
    assert network.shape[0]==network.shape[1]
    network = network[network.index]
    network = network.fillna(-666) #missing value
    network[network==0] = 2 #means that the edge is present both as activatory and inhibitory
    network[network==-666] = 0
    network[network<0] = -1 #only inhibitory
    network[(network>0)&(network<2)] = 1 #only activatory
    network = network.astype(int)
    network_unsigned, network_signed = [deepcopy(network) for _ in range(2)] 
    network_unsigned[network_unsigned!=2] = 0
    network_unsigned = network_unsigned/2 #1 if unsigned edge exists, 0 otherwise
    network_signed[network_signed==2] = 0 #1 if signed activatory edge exists, -1 if signed inhibitory edge exists, 0 otherwise
    # Correlation matrix
    if (expr_df is not None):
        df = quantile_normalize(expr_df)
        df = df.loc[[g for g in list(network.index) if (g in df.index)]]
        coexpr = np.power(df.T.corr(method=cor_method), beta)
        coexpr[coexpr.abs()<tau] = 0
        if (accept_nonRNA):
            missing_genes = [g for g in list(network.index) if (g not in df.index)]
            for g in missing_genes:
                coexpr[g] = 1
                coexpr.loc[g] = 1
            coexpr = coexpr.loc[list(network.index)][list(network.index)]
    else:
        coexpr = pd.DataFrame(np.ones(network.shape), index=list(network.index), columns=list(network.index)) #default: positive interactions
    assert all([coexpr.shape[i]==s for i,s in enumerate(network_unsigned.shape)])
    assert all([coexpr.index[i]==s for i,s in enumerate(network_unsigned.index)])
    assert all([coexpr.columns[i]==s for i,s in enumerate(network_unsigned.columns)])
    net_mat = np.multiply((-1)**(coexpr.values<0).astype(int), network_unsigned.values).astype(float)#equal to 1 if strong correlation and activatory, -1 if strong correlation and inhibitory, 0 otherwise
    influences = pd.DataFrame(net_mat+network_signed.values, index=network.index, columns=network.columns)
    assert not pd.isnull(influences).any().any()
    return influences

def create_grn(influences, exact=False, max_maxclause=3, quiet=False):
    '''
    Create a BoneSiS InfluenceGraph

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes] of interactions, values in {-1,1,0} -1:negative,1:positive,0:absent
    exact : Python bool
        [default=False] : should all interactions be preserved?
    max_maxclause : Python integer
        [default=3] : upper bound on the number of clauses in DNF form
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    grn : BoneSiS InfluenceGraph class object
        BoneSiS GRN object
    '''
    assert all([n in [-1,1,0] for n in np.unique(influences.values)])
    maxclause = min(get_maxdegree(influences, quiet=quiet), max_maxclause)
    influences_list = influences.melt(ignore_index=False)
    influences_list["id"] = influences_list.index
    influences_list = influences_list[influences_list["value"]!=0]
    influences_list = [(x,y,dict(sign=s)) for [y,s,x] in influences_list.values.tolist()]
    if (not quiet):
        print("<BONESIS> %d interactions (maximum # of clauses = %d)" % (len(influences_list), maxclause))
    grn = bonesis.InfluenceGraph(graph=influences_list, maxclause=maxclause, exact=exact)
    return grn

def get_maxdegree(influences, activatory=True, quiet=False):
    '''
    Computes the maximum ingoing degree (or the maximum number of potential activatory regulators) in a graph

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes] of interactions: -1:negative,1:positive,0:absent
    activatory : Python bool
        [default=True] : computes the maximum number of potential activatory regulators instead
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    maxindegree : Python integer
        maximum ingoing degree (or the maximum number of potential activatory regulators)
    '''
    assert all([n in [-1,0,1] for n in np.unique(influences.values)])
    if (not activatory):
        maxindegree=int(np.max((influences!=0).astype(float).sum(axis=0)))
    else:
        maxindegree=int(np.max((influences>0).astype(float).sum(axis=0)))
    if (not quiet):
        print("<UTILS_GRN> Maximum "+("ingoing degree" if (not activatory) else "possible #activators")+"=%d" % maxindegree)
    return maxindegree

def build_observations(grn, signatures, quiet=False):
    '''
    Implement experimental constraints from perturbation experiments in signatures. Experimental constraints are of the form Initial state masked by single-gene perturbation can lead to a steady attractor state Final

    ...

    Parameters
    ----------
    grn : InfluenceGraph (from BoneSiS)
        contains topological constraints
    signatures : Pandas DataFrame
        rows/[genes] x columns/[experiment IDs]. Experiment IDs is of the form "<pert. gene>_<pert. type>_<...>_<cell line>" (treated) or "initial_<cell line>" (control)
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    BO : BoNesis object (from BoneSiS)
        BoNesis object which can be evaluated
    '''
    data_exps = {}
    if (len(signatures)==0):
        BO = bonesis.BoNesis(grn, data_exps)
        return BO
    signatures = signatures.loc[[g for g in signatures.index if (g in grn.nodes)]]
    ## 1. Add signatures of experimental states
    exps = [x for x in signatures.columns if ("initial" not in x)]
    exps_ids = range(len(exps))
    ## 1a. Instantiate experiments
    ## zero state to ignore
    #data_exps.update({"zero": {g: 0 for g in grn.nodes}})
    ## 1b. For each experiment (1 initial state, 1 final state, 1 perturbation)
    for exp_nb in exps_ids:
        cell = exps[exp_nb].split("_")[-1]
        cols = ["Exp%d_"%(exp_nb+1)+x for x in ["init", "final"]]
        data_df = signatures[["initial_"+cell, exps[exp_nb-1]]]

        ## Compatible with perturbation experiment
        data_df = data_df.T
        pert, sign = exps[exp_nb].split("_")[0], 0 if (exps[exp_nb].split("_")[1]=="KD") else 1
        data_df[pert] = sign 
        data_df = data_df.T

        data_df.columns = cols
        for col in cols:
            data_exps.update(data_df[[col]].dropna().astype(int).to_dict())
    if (not quiet):
        print_exps = pd.DataFrame.from_dict(data_exps, orient="index").fillna(-1).astype(int)
        print_exps[print_exps==-1] = ""
        print("\n<UTILS_GRN> %d experiments\n%s" % (len(exps_ids), str(print_exps)))
    BO = bonesis.BoNesis(grn, data_exps)
    ## 2. Instantiate reachability & fixed point constraints
    for exp_nb in exps_ids:
        exp = exps[exp_nb]
        pert, sign = exp.split("_")[:2]
        cell = exp.split("_")[-1]
        sign = int(sign!="KD")
        with BO.mutant({pert: sign}) as m:
            final_FP = m.fixed(~m.obs("Exp%d_final" % (exp_nb+1)))
            ## 2a. There exists a trajectory : init -> trapspace
            ~m.obs("Exp%d_init" % (exp_nb+1)) >= final_FP #~m.obs("Exp%d_final" % (exp_nb+1)) ##
            ## MEMO: Universal reachable fixed points: ~m.obs("Exp%d_init" % (exp_nb+1)) >> "fixpoints" ^ {m.obs("Exp%d_final" % (exp_nb+1))}
        ## 2b. No trivial trajectory from zero state : zero -> trapspace(unless the experiment actually involves it)
        #state_di = data_exps["Exp%d_init" % (exp_nb+1)]
        #state_di.update({pert: sign})
        #if (all([state_di.get(g, 1)==0 for g in grn.nodes])):
        #    continue
        #~BO.obs("zero") / ~BO.obs("Exp%d_final" % (exp_nb+1))
    ## MEMO: Fixed point: BO.fixed(~BO.obs("attractor_%d" % (state_id+1)))
    ## MEMO: Trap space: BO.fixed(BO.obs("attractor_%d" % (state_id+1)))
    return BO

def solution2influences(solution):
    '''
    Converts a solution object into a influences object

    ...

    Parameters
    ----------
    solution : Pandas Series
        rows/[genes]

    Returns
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes] contains values {-1,1,0,2} -1: negative, 1: positive, 0: absent, 2: non monotonic
    '''
    gene_list = list(solution.index)
    N = len(gene_list)
    infl_mat = np.zeros((N,N))
    grfs = get_grfs_from_solution(solution)
    for gene in grfs:
        ig = gene_list.index(gene)
        for regulator in grfs[gene]:
            if (regulator in gene_list):
                ir = gene_list.index(regulator)
                infl_mat[ir,ig] = grfs[gene][regulator]
            else:
                infl_mat = np.concatenate((infl_mat, np.zeros((1,N))), axis=0)
                infl_mat = np.concatenate((infl_mat, np.zeros((N+1,1))), axis=1)
                infl_mat[N,ig] = grfs[gene][regulator]
                gene_list.append(regulator)
                N += 1
    influences = pd.DataFrame(infl_mat, index=gene_list, columns=gene_list)
    return influences

def zip2df(fname):
    '''
    Extract solutions in ZIP file as DataFrames

    ...

    Parameters
    ----------
    fname : Python character string
        zip file which contains BNET solutions

    Returns
    ----------
    solutions : Pandas DataFrame
        rows/[genes] x columns/[solutions] the GRFs for each gene in each solution
    '''
    with ZipFile(fname, "r") as zip:
        zip.extractall()
    grn_fnames = glob("*.bnet")
    grns = []
    for grn_fname in grn_fnames:
        grn = load_grn(grn_fname)
        grns.append(grn)
    solutions = pd.DataFrame(grns)
    sbcall("rm -f *.bnet", shell=True)
    return solutions

def load_grn(fname):
    '''
    Loads GRN as MPBN class element

    ...

    Parameters
    ----------
    fname : Python character string
        BNET file

    Returns
    ----------
    BN : mpbn.MPBooleanNetwork object
        Boolean network with Most Permissive semantics
    '''
    BN = mpbn.MPBooleanNetwork(fname)
    sep=" <- " if (" <- " in str(BN)) else ", "
    solution = pd.Series([x.split(sep)[-1] for x in str(BN).split("\n")], index=[x.split(sep)[0] for x in str(BN).split("\n")])
    return solution

def get_minimal_edges(R, maximal=False):
    '''
    Return one of the solutions with the smallest (or greatest) number of edges

    ...

    Parameters
    ----------
    R : Pandas DataFrame
        rows/[genes] x columns/[solution IDs]
    connected : Python bool
        [default=False] : if set to True, return the CONNECTED solution which satisfies those constraints
    maximal : Python bool
        [default=False] : if set to True, return the solution with the greatest number of edges

    Returns
    ----------
    solution, nedges : Python integer x Python integer
        solution and corresponding number of edges
    '''
    nedges_list = [int(solution2influences(R[col]).abs().sum().sum()) for col in R.columns]
    solution_id = (np.argmax if (maximal) else np.argmin)(nedges_list)
    return R[R.columns[solution_id]], nedges_list[solution_id]

def get_grfs_from_solution(solution):
    '''
    Retrieve all gene regulatory functions (GRFs) from a given solution

    ...

    Parameters
    ----------
    solution : Pandas Series
        rows/[genes] 

    Returns
    ----------
    grfs : Python dictionary
        {gene: {regulator: sign, ...}, ...} where sign in {-1,1} -1: inhibitor, 1: activator
    '''
    sol_dict, grfs = solution.to_dict(), {}
    for gene in sol_dict:
        grf = "".join("".join(" ".join(" ".join(str(sol_dict[gene]).split(", ")[-1].split("|")).split("&")).split("(")).split(")"))
        if (len(str(grf))==0):
            continue
        regulators = {}
        if (str(grf) not in ["0", "1"]):
            reg_names = list(set(grf.split(" ")))
            regulators_df = pd.DataFrame([[(-1)**int(regulator[0]=="!")] for regulator in reg_names], index=[regulator[int(regulator[0]=="!"):] for regulator in reg_names], columns=["regulators"])
            regulators_df = regulators_df.groupby(level=0).sum()
            regulators_df[regulators_df==0] = 2
            regulators = regulators_df.to_dict()["regulators"]
        grfs.setdefault(gene, regulators)
    return grfs

def save_grn(solution, fname, sep=", ", quiet=False, max_show=5, write=True):
    '''
    Write and/or print .bnet file

    ...

    Parameters
    ----------
    solution : Pandas Series
        rows/[genes] contains gene regulatory functions (GRF)
    fname : Python character string
        where to write the file (w/o .bnet extension)
    sep : Python character string
        what separates regulators from regulated genes
    quiet : Python bool
        [default=False] : prints out verbose
    max_show : Python integer
        [default=5] : maximum number of printed GRFs
    write : Python bool
        [default=True] : if set to True, write to a .bnet file

    Returns
    ----------
    None
        writes the GRN to a file fname
    '''
    sol = solution.to_dict()
    print_sol = ["".join([str(x) for x in [gene, sep, sol[gene]]]) for gene in sol if (len(gene)>0)]
    if (not quiet):
        print("\n"+("\n".join(print_sol[:max_show]+["..." if (len(print_sol)>max_show) else ""])))
    print_sol = "\n".join(print_sol)
    if (write):
        with open(fname+".bnet", "w+") as f:
            f.write(print_sol)
    return None

def save_solutions(bnetworks, fname, limit):
    '''
    Enumerate and save solutions

    ...

    Parameters
    ----------
    bnetworks : Bonesis object
        Output of the inference
    fname : Python character string
        ZIP filename to store the solutions
    limit : Python integer
        maximum number of solutions to enumerate

    Parameters
    ----------
    n : Python integer
        number of enumerated solutions
    '''
    with ZipFile(fname, "w") as bundle:
        n = 0
        for i, bn in enumerate(bnetworks):
            with bundle.open(f"bn{i}.bnet", "w") as fp:
                fp.write(bn.source().encode())
            n += 1
            if n == limit:
                break
    return n

def infer_network(BO, njobs=1, fname="solutions", use_diverse=True, limit=50, niterations=1):
    '''
    Infer solutions matching topological & experimental constraints

    ...

    Parameters
    ----------
    BO : Bonesis object (from BoneSiS)
        contains topological & experimental constraints
    fname : Python character string
        [default="solutions"] : path to solution files
    use_diverse : Python bool
        [default=True] : use the "diverse" procedure in BoneSiS
    limit : Python integer
        [default=50] : maximum number of solutions to generate per interation
    niterations : Python integer
        [default=1] : maximum number of iterations

    Returns
    ----------
    nsolutions : Python integer
        list of # solutions per iteration
    '''
    bonesis.settings["parallel"] = njobs
    #bonesis.settings["solutions"] = "subset-minimal"
    if (use_diverse):
        infer = lambda bo : bo.diverse_boolean_networks
    else:
        infer = lambda bo : bo.boolean_networks
    param_di = {"skip_supersets": True}
    if (str(limit)!="None"):
        param_di.setdefault("limit",limit)
    bnetworks = infer(BO)(**param_di)
    nsolutions = []
    for niter in tqdm(range(niterations)):
        if (not os.path.exists(fname+"_"+str(niter+1)+".zip")):
            nsolutions.append(save_solutions(bnetworks, fname+"_"+str(niter+1)+".zip", limit))
    return nsolutions

def get_genes_downstream(network_fname, gene, n=-1):
    '''
    Get the list of genes downstream of a gene in a network

    ...

    Parameters
    ----------
    network_fname : Python character string
        path to the .BNET file associated with the network
    gene : Python character string
        gene name in the network
    n : Python integer
        [default=-1] : number of recursions (if<0, recursively get all downstream genes)

    Returns
    ----------
    lst_downstream : Python character string list
        list of nodes downstream of @gene
    '''   
    with open(network_fname, "r") as f:
        grf_list = f.read().split("\n")
    grfs = dict([x.split(", ") for x in grf_list if (len(x)>0)]) 
    set_downstream = set([])
    if (gene not in grfs):
        return []
    count_n = 0
    while ((n<0) or (count_n<n)):
        l = len(set_downstream)
        for g in grfs:
            if (any(x in grfs[g] for x in set_downstream.union({gene}))):
                set_downstream = set_downstream.union({g})
        if (l==len(set_downstream)):
            break
        count_n += 1
    return list(set_downstream)

def get_genes_most_variable(control_profiles, treated_profiles, p=0.8):
    '''
    Get the list of genes which contribute most to the variation between two conditions (in the @pth percentile of change)

    ...

    Parameters
    ----------
    control_profiles : Pandas DataFrame
        rows/[genes] x columns/[samples] profiles from condition 1
    treated_profiles : Pandas DataFrame
        rows/[genes] x columns/[samples] profiles from condition 1
    p : Python float
        100*p th percentile to consider 

    Returns
    ----------
    lst_genes : Python character string list
        list of nodes which contribute most to the variation between conditions
    '''  
    assert p>=0 and p<=1
    model = Logit(penalty="l1",solver="saga",fit_intercept=False,random_state=0,max_iter=1000,n_jobs=njobs)
    df = control_profiles.join(treated_profiles, how="inner").fillna(0.5)
    X = df.T.values
    y = np.ravel(np.array([0]*control_profiles.shape[1]+[1]*treated_profiles.shape[1]))
    model.fit(X, y)
    q = np.quantile(np.abs(model.coef_.flatten()), p)
    genes = {g: model.coef_[0,ig] for ig, g in enumerate(list(df.index)) if (abs(model.coef_[0,ig])>q)}
    return list(genes.keys())

def reconnect_network(network_fname):
    '''
    Write the network with all isolated nodes (no ingoing/outgoing edges) filtered out

    ...

    Parameters
    ----------
    network_fname : Python character string
        path to the .BNET associated with the network

    Returns
    ----------
    fname : Python character string
        path to the .BNET associated with the reconnected network
    '''  
    with open(network_fname, "r") as f:
        network = pd.DataFrame({"Solution": dict([["_".join(g.split("-")) for g in x.split(", ")] for x in f.read().split("\n") if (len(x)>0)])})
    influences = solution2influences(network["Solution"])
    assert influences.shape[0]==influences.shape[1]
    assert all([influences.index[i]==influences.columns[i] for i in range(influences.shape[0])])
    influences = influences.loc[(influences.abs().sum(axis=1)>0)&(influences.abs().sum(axis=0)>0)]
    gene_list = list(influences.index)
    network_connected = network.loc[gene_list]
    def get_all_genes(network):
        all_genes = list(network.index)
        for idx in network.index:
            grf = str(network.loc[idx]["Solution"])
            for symb in ["!","&","|","(",")"]:
                grf = " ".join(grf.split(symb))
            gene_lst = [g for g in grf.split(" ") if (g not in [" ", "0","1"] and len(g)>0)]
            all_genes += gene_lst
        return list(set(all_genes))
    full_lst = get_all_genes(network_connected)
    network_connected = network.loc[full_lst].to_dict()["Solution"]
    network_connected = "\n".join([", ".join(["_".join(g.split("-")) for g in [x,y]]) for x,y in network_connected.items()])
    network_fname_connected = network_fname.split(".bnet")[0]+"_connected.bnet"
    with open(network_fname_connected, "w") as f:
        f.write(network_connected) 
    return network_fname_connected



#coding: utf-8

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
import seaborn as sns

def influences2graph(influences, fname, optional=False, compile2png=True, engine=["sfdp","dot"][0]):
    '''
    Plots a network by conversion to a DOT file and then to PNG

    ...

    Parameters
    ----------
    influences : Pandas DataFrame
        rows/[genes] x columns/[genes], contains {-1,1,2}
    fname : Python character string
        filename of png file
    optional : Python bool
        [default=False] : should interactions be drawn as optional (dashed lines)?

    Returns
    ----------
    None
        writes a DOT file which can be converted to PNG image (if compile2png=True)
    '''
    dotfile = fname+".dot"
    filename = fname+".png"
    graph = ["digraph {"]
    for idx in influences.columns:
        if (len(idx)>0):
            graph += ["\""+idx+"\" [shape=circle,fillcolor=grey,style=filled];"]
    directed, undirected = ["subgraph D {"], ["subgraph U {"]
    in_undirected = []
    for x,y in np.argwhere(influences.values!=0).tolist():
        nx, ny = influences.index[x], influences.columns[y]
        sign = influences.values[x,y]
        if ((y,x,sign) in in_undirected):
            continue
        edge = " ".join(["\""+nx+"\"", "->", "\""+ny+"\""])
        edge += " ["
        if (influences.values[y,x]!=0):
            edge += "dir=none,"
        edge += "label="+("\"-\"" if (sign<0) else ("\"+\"" if ((sign>0) and (sign<2)) else "\"+-\""))
        edge += ",penwidth=3,color="+("red" if (sign<0) else ("green" if ((sign>0) and (sign<2)) else "black"))
        edge += ",arrowhead=tee" if (sign < 0) else ""
        edge += ",style=dashed" if (optional) else ""
        edge += "];"
        if (influences.values[y,x]!=0):
            undirected += [edge]
            in_undirected += [(x,y,sign)]
        else:
            directed += [edge]
    directed += ["}"]
    undirected += ["}"]
    graph += (directed if (len(directed)>2) else [])+(undirected if (len(undirected)>2) else [])+["}"]
    with open(dotfile, "w+") as f:
        f.write("\n".join(graph))
    if (compile2png):
        from subprocess import call as sbcall
        sbcall(engine+" -Tpng "+dotfile+" > "+filename+" && rm -f "+dotfile, shell=True)
    return None

def plot_influence_graph(network_df, input_col, output_col, sign_col, direction_col=None, fname="graph.png", optional=True):
    '''
    Converts a network into a PNG picture

    ...

    Parameters
    ----------
    network_df : Pandas DataFrame
        rows/[index] x columns/[input_col,output_col,sign_col]
    input_col,output_col,sign_col,direction_col : Python character string
        columns of network_df
    fname : Python character string
        [default="graph.png"] : file name for PNG picture
    optional : Python bool
        [default=True] : should edges be plotted as dashed lines?

    Returns
    ----------
    None
        Creates a image of the graph in file fname
    '''
    influences = network_df.pivot_table(index=input_col, columns=output_col, values=sign_col, aggfunc='mean')
    influences[influences==0] = 2
    influences = influences.fillna(0)
    influences[influences<0] = -1
    influences[(influences>0)&(influences<2)] = 1
    missing_cols = [g for g in influences.index if (g not in influences.columns)]
    influences = pd.concat((influences, pd.DataFrame([], index=influences.index, columns=missing_cols).fillna(0)), axis=1)
    missing_idxs = [g for g in influences.columns if (g not in influences.index)]
    influences = pd.concat((influences, pd.DataFrame([], index=missing_idxs, columns=influences.columns).fillna(0)), axis=0)
    infl_mat = influences.values
    if (direction_col is not None):
        for x in network_df.index:
            is_directed = bool(network_df.loc[x][direction_col])
            if (not is_directed):
                inp = list(influences.index).index(str(network_df.loc[x]["preferredName_A"]))
                out = list(influences.columns).index(str(network_df.loc[x]["preferredName_B"]))
                infl_mat[inp, out] = int(network_df.loc[x][sign_col])
                infl_mat[out, inp] = infl_mat[inp, out]
    influences = pd.DataFrame(infl_mat, index=influences.index, columns=influences.columns).loc[influences.index][influences.index]
    influences = influences.astype(int)
    influences.index = influences.index.astype(str)
    influences.columns = influences.columns.astype(str)
    influences2graph(influences, fname, optional=optional)
    return None

def plot_signatures(signatures, perturbed_genes=None, width=10, height=10, max_show=50, fname="signatures"):
    '''
    Print signatures

    ...

    Parameters
    ----------
    signatures : Pandas DataFrame
        rows/[genes] x columns/[signature IDs]
    perturbed_genes : Python character string list
        [default=None] : list of gene names perturbed in the signatures
    width, height : Python integer
        [default=10] : dimensions of image
    max_show : Python integer
        [default=50] : maximum number of genes shown (as only the @max_show genes with highest variance across signatures are plotted)
    fname : Python character string
        [default="signatures"] : path of resulting PNG image

    Returns
    ----------
    None
        plots the signatures as heatmaps in file fname
    '''
    from matplotlib import colors,rc
    rc("ytick", labelsize=5)
    rc("xtick", labelsize=10)
    sigs = deepcopy(signatures)
    sigs[sigs == 0] = -1
    sigs[pd.isnull(sigs)] = 0 
    max_genes = np.argsort(np.var(sigs.values, axis=1))[-max_show:]
    max_genes = sigs.index[max_genes]
    for g in perturbed_genes:
        if (g not in sigs.index):
            sigs.loc[g] = 0
    max_genes = perturbed_genes+[g for g in max_genes if (g not in perturbed_genes)]
    sigs_ = sigs.loc[max_genes]
    fig, ax = plt.subplots(figsize=(width,height))
    cmap = colors.ListedColormap(['red', 'black', 'green'])
    bounds=[-1, -0.5, 0.5, 1]
    norm = colors.BoundaryNorm(bounds, cmap.N)
    pos = ax.imshow(sigs_, cmap=cmap, origin="lower", interpolation='nearest', norm=norm) #remove score line
    plt.colorbar(pos, ax=ax, ticks=[-1, 0, 1], shrink=0.12, aspect=4)
    ax.set_xticks(range(sigs_.shape[1]))
    ax.set_xticklabels(sigs.columns, rotation=90)
    ax.set_yticks(range(sigs_.shape[0]))
    ax.set_yticklabels(sigs_.index)
    plt.savefig(fname+".png", bbox_inches="tight")
    plt.close()
    return None

def plot_distributions(profiles, fname="gene_expression_distribution.png", thres=None):
    '''
    Plots the distributions (boxplots) of gene expression across samples for each gene, and the selected threshold for binarization

    ...

    Parameters
    ----------
    profiles : Pandas DataFrame
        rows/[genes+annotations] x columns/[samples]
    fname : Python character string
        [default="gene_expression_distribution.png"] : file name
    thres : Python float or None
        [default=None] : binarization threshold (if there is any)


    Returns
    ----------
    None
        plots boxplots of expression for each gene in profiles
    '''
    add_rows_profiles = ["annotation", "perturbed", "perturbation", "cell_line", "sigid"]
    bp = profiles.loc[[g for g in profiles.index if (g not in add_rows_profiles)]].T.apply(pd.to_numeric).boxplot(rot=90, figsize=(25,15))
    if (str(thres)!="None"):
        K = profiles.shape[0]-3
        for t in [thres, 1-thres]:
            plt.plot(list(range(K)), [t]*K, "r--")
    plt.xlabel("Genes")
    plt.ylabel("Gene expression")
    plt.savefig(fname, bbox_inches="tight")
    plt.close()
    return None

def plot_discrete_distributions(signatures, fname="signature_expression_distribution.png"):
    '''
    Plots the distributions (histograms) of genes with determined status across signatures

    ...

    Parameters
    ----------
    signatures : Pandas DataFrame
        rows/[genes] x columns/[samples] with values in {0,NaN,1}. Determined status is either 0 or 1.
    fname : Python character string
        [default="signature_expression_distribution.png"] : file name

    Returns
    ----------
    None
        plots the number of genes with expression values 0, 1 or NaN in each signature
    '''
    N = int(np.sqrt(signatures.shape[1]))+1
    fig, axes = plt.subplots(nrows=N, ncols=N, figsize=(25,15))
    for iax in range(signatures.shape[1]):
        i,j = iax//N, iax%N
        sig = signatures.iloc[:, iax].fillna(0.5)
        axes[i,j].set_xticks([0,0.5,1])
        axes[i,j].set_xticklabels(('0','NaN','1'))
        axes[i,j].hist(sig.values.T)
        axes[i,j].set_ylabel("#genes")
        axes[i,j].set_title(signatures.columns[iax])
    plt.xlabel("Gene expression level")
    plt.ylabel("# genes")
    plt.savefig(fname, bbox_inches="tight")
    plt.close()
    return None

def plot_boxplots(scores, patient_scores, ground_truth=None, fsize=12, msize=5, fname="boxplots.pdf"):
    '''
    Plots one boxplot per treatment (all values obtained on patient profiles)

    ...

    Parameters
    ----------
    scores : Pandas DataFrame
        rows/[drug names] x column/[value]
    patient_scores : Pandas DataFrame
        rows/[drug names] x columns/[patient samples]
    ground_truth : Pandas DataFrame
        [default=None] : rows/[drug names] x column/[class] Values in 1: treatment, 0: unknown, -1: aggravating. If not provided: does not color boxplots according to the class
    fsize : Python integer
        [default=18] : font size
    msize : Python integer
        [default=5] : marker size
    fname : Python character string
        [default="boxplots"] : file name for the plot

    Returns
    ----------
    None
        create boxplots of reward scores across patients for each drug
    '''
    scores_ = deepcopy(scores).sort_values(by=scores.columns[0], ascending=False)
    drug_names = list(scores_.index)
    sorted_rewards_list = list(scores_[scores_.columns[0]])
    sorted_rewards = patient_scores.loc[drug_names,:]
    if (ground_truth is not None):
        gt_scores = ground_truth.loc[[t for t in scores_.index if (t in ground_truth.index)]]
        gt_missing = pd.DataFrame([], index=[t for t in scores_.index if (t not in ground_truth.index)], columns=ground_truth.columns)
        gt_scores = pd.concat((gt_scores, gt_missing), axis=0).fillna(0).loc[drug_names].astype(int)
        sorted_ground_truth = list(gt_scores[gt_scores.columns[0]])
        positive = [si+1 for si, s in enumerate(sorted_ground_truth) if (s > 0)]
        negative = [si+1 for si, s in enumerate(sorted_ground_truth) if (s < 0)]
    else:
        positive, negative = [], []
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10,6))
    bplot = sns.boxplot(sorted_rewards.T, showmeans=False, meanline=False, notch=False, ax=ax)
    if (ground_truth is not None):
        colors = {1: "green", -1: "red", 0: "white"}
        for i in range(len(sorted_rewards)):
            mybox = bplot.artists[i]
            mybox.set_facecolor(colors[sorted_ground_truth[i]])
    ax.plot(range(len(drug_names)), sorted_rewards_list, "kD", label='score')
    ax.set_ylabel("Scores across patient profiles", fontsize=fsize)
    ax.set_xlabel("Drug", fontsize=fsize)
    ax.set_xticklabels(["%s (%.2f)" % (drug_names[i], r) for i, r in enumerate(sorted_rewards_list)], rotation=90, fontsize=fsize)
    plt.yticks(rotation=0)
    cols=["r" if (i in negative) else ("g" if (i in positive) else "k") for i in range(1, scores.shape[0]+1)]
    for xtick, color in zip(ax.get_xticklabels(), cols):
        xtick.set_color(color)
    ax.tick_params(axis='y', which='major', labelsize=fsize)
    plt.legend(fontsize=fsize)
    plt.savefig(fname, bbox_inches="tight")
    plt.close()

def plot_heatmap(X, ground_truth=None, fname="heatmap.pdf", w=20, h=20, bfsize=20, fsize=20, rot=75):
    '''
    Plots an heatmap of the signatures, with the potential ground truth

    ...

    Parameters
    ----------
    X : Pandas DataFrame
        rows/[features] x columns/[drug names]
    ground_truth : Pandas DataFrame
        [default=None] : rows/[drug names] x column/[class] Values in 1: treatment, 0: unknown, -1: aggravating. If not provided: does not color boxplots according to the class
    fname : Python character string
        [default="heatmap.pdf"] : file name for the plot
    w : Python integer
        [default=20] : figure width
    h : Python integer
        [default=20] : figure height
    bfsize : Python integer
        [default=20] : font size in the color bar
    rot : Python integer
        [default=75] : rotation angle of labels

    Returns
    ----------
    None
        plots an heatmap of similarity across drugs based on the Pearson correlation
    '''
    colors = {1: "green", -1: "red", 0: "black"}
    if (ground_truth is not None):
        gt_scores = ground_truth.loc[[t for t in X.columns if (t in ground_truth.index)]]
        gt_missing = pd.DataFrame([], index=[t for t in X.columns if (t not in ground_truth.index)], columns=ground_truth.columns)
        gt_scores = pd.concat((gt_scores, gt_missing), axis=0).fillna(0).loc[X.columns].astype(int)
    else:
        gt_scores = pd.DataFrame([], index=X.columns, columns=["Ground Truth"]).fillna(0)
    row_colors = pd.Series([colors[i] for i in list(gt_scores[gt_scores.columns[0]])], index=list(gt_scores.index))
    im = sns.clustermap(X.corr(method="pearson"), cmap="Blues", figsize=(w,h), row_cluster=True, col_cluster=True, 
	row_colors=row_colors, col_colors=row_colors, vmin=-1, vmax=1)
    im.ax_heatmap.axes.set_yticklabels([t.get_text() for t in im.ax_heatmap.axes.get_yticklabels()], fontsize=fsize)
    im.ax_heatmap.axes.set_xticklabels([t.get_text() for t in im.ax_heatmap.axes.get_xticklabels()], fontsize=fsize, rotation=rot)
    for ti, tick_label in enumerate(im.ax_heatmap.axes.get_yticklabels()):
        tick_label.set_color(colors[gt_scores.loc[tick_label.get_text()][gt_scores.columns[0]]])
    for ti, tick_label in enumerate(im.ax_heatmap.axes.get_xticklabels()):
        tick_label.set_color(colors[gt_scores.loc[tick_label.get_text()][gt_scores.columns[0]]])
    cbar = im.ax_heatmap.axes.collections[0].colorbar
    cbar.ax.tick_params(labelsize=bfsize)
    plt.savefig(fname, bbox_inches="tight")
    plt.close()

def plot_roc_curve(pr, prs, tr, fname="ROC.pdf", method_name="predictor", fsize=18):
    '''
    Plots a ROC curve (with variations across samples)

    ...

    Parameters
    ----------
    pr : Pandas DataFrame
        rows/[drug names] x column/[value]
    prs : Pandas DataFrame
        rows/[drug names] x columns/[patient samples]
    tr : Pandas DataFrame
        [default=None] : rows/[drug names] x column/[class] 
    fname : Python character string
        [default="ROC.pdf"] : file name for the plot
    method_name : Python character string
        [default="predictor"] : name of the predictor
    fsize : Python integer
        [default=18] : font size


    Returns
    ----------
    None
        Plots a ROC curve based on the drug repurposing predictions
    '''
    from sklearn.metrics import roc_curve, roc_auc_score
    from scipy import interp
    predicted = pr.values.tolist()
    truth = tr.loc[pr.index].values.tolist()
    ids = np.argsort(predicted).tolist()
    ids.reverse()
    sorted_rewards = [predicted[i] for i in ids]
    sorted_ground_truth = [int(truth[i]>0) for i in ids]
    AUC = roc_auc_score(sorted_ground_truth, sorted_rewards)
    base = np.linspace(0, 1, 101)
    tprs, tprs2 = [], []
    aucs = []
    for idx in prs.columns:
        predicted_patient = prs.loc[pr.index][idx].values.tolist()
        fper, tper, _ = roc_curve(truth, predicted_patient)
        aucs.append(roc_auc_score(truth, predicted_patient))
        tpr = interp(base, fper, tper)
        tpr[0] = 0.0
        tprs.append(tpr)
    tprs = np.array(tprs)
    mean_tprs = tprs.mean(axis=0)
    std_tprs = tprs.std(axis=0)
    tprs_upper = np.minimum(mean_tprs + std_tprs, 1)
    tprs_lower = np.maximum(mean_tprs - std_tprs, 0)
    plt.plot(base, mean_tprs, 'r', alpha = 0.8, label="%s (AUC=%.2f)" % (method_name.split("_")[0],AUC))
    plt.fill_between(base, tprs_lower, tprs_upper, color = 'red', alpha = 0.2)
    plt.plot([0, 1], [0, 1], color='blue', lw=2, linestyle='--', alpha=0.8)
    plt.xticks(fontsize=fsize)
    plt.yticks(fontsize=fsize)
    plt.xlabel('False Positive Rate', fontsize=fsize)
    plt.ylabel('True Positive Rate', fontsize=fsize)
    plt.title('Receiver Operating Characteristic Curve', size=fsize)
    plt.legend(loc="lower right", fontsize=fsize)
    plt.savefig(fname, bbox_inches="tight")
    plt.close()

def plot_precision_recall(pr, prs, tr, beta=1, thres=0.5, fname="PRC.pdf", method_name="predictor", fsize=18):
    '''
    Plots a Precision-Recall curve (with variations across samples)

    ...

    Parameters
    ----------
    pr : Pandas DataFrame
        rows/[drug names] x column/[value]
    prs : Pandas DataFrame
        rows/[drug names] x columns/[patient samples]
    tr : Pandas DataFrame
        [default=None] : rows/[drug names] x column/[class] 
    beta : Python float
        [default=1] : value of coefficient beta for the F-measure
    thres : Python float
        [default=0.5] : decision threshold
    fname : Python character string
        [default="PRC.pdf"] : file name for the plot
    method_name : Python character string
        [default="predictor"] : name of the predictor
    fsize : Python integer
        [default=18] : font size


    Returns
    ----------
    None
        Plots a Precision-Recall curve based on the drug repurposing predictions
    '''
    from sklearn.metrics import precision_recall_curve, fbeta_score
    from scipy import interp
    predicted = pr.values.tolist()
    truth = tr.loc[pr.index].values.tolist()
    ids = np.argsort(predicted).tolist()
    ids.reverse()
    sorted_rewards = [int(predicted[i]>thres) for i in ids]
    sorted_ground_truth = [int(truth[i]>0) for i in ids]
    f = fbeta_score(sorted_ground_truth, sorted_rewards, average='weighted', beta=beta)
    base = np.linspace(0, 1, 101)
    recs, recs2 = [], []
    for idx in prs.columns:
        predicted_patient = prs.loc[pr.index][idx].values.tolist()
        pres, rec, _ = precision_recall_curve(truth, predicted_patient)
        rec = interp(base, pres, rec)
        recs.append(rec)
    recs = np.array(recs)
    mean_recs = recs.mean(axis=0)
    std_recs = recs.std(axis=0)
    recs_upper = np.minimum(mean_recs + std_recs, 1)
    recs_lower = np.maximum(mean_recs - std_recs, 0)
    plt.plot(base, mean_recs, 'b', alpha=0.8, label=r'%s (F_%.1f = %.2f)' % (method_name.split("_")[0], beta, f))
    plt.fill_between(base, recs_lower, recs_upper, color="blue", alpha=0.2)
    plt.plot([0,1], [1,0], linestyle="--", lw=2, color="blue", alpha=0.8)
    plt.xticks(fontsize=fsize)
    plt.yticks(fontsize=fsize)
    plt.xlabel('Precision', fontsize=fsize)
    plt.ylabel('Recall', fontsize=fsize)
    plt.title('Precision-Recall Curve', fontsize=fsize)
    plt.legend(loc='lower left', fontsize=fsize)
    plt.savefig(fname,bbox_inches="tight")
    plt.close()



#coding: utf-8

import NORDic
import os
import pandas as pd
from multiprocessing import cpu_count
import numpy as np
import unittest

## IV. Drug Repurposing (**NORDic DR**)

class TestDR(unittest.TestCase):

    def create_instance(self):
        file_folder="ToyOndine/"
        seed_number=12345
        njobs=min(5,max(1,cpu_count()-2))
        SIMU_params = {'nb_sims': 1000, 'rates': "fully_asynchronous", 'thread_count': njobs, 'depth': "constant_unitary"}

        np.random.seed(seed_number)
        with open(file_folder+"solution.bnet", "r") as f:
            genes = [line.split(", ")[0] for line in f.read().split("\n") if (len(line)>0)]

        ## Create a set of random "patient" and "control" states (see tests_PMR.py)
        states = pd.read_csv(file_folder+"phenotypes.csv", index_col=0)
        ### Patient profiles
        patients = states[[c for c in states.columns if ("Case" in c)]].loc[[i for i in states.index if (i != "annotation")]]

        ## Drug targets
        targets = pd.read_csv(file_folder+"targets.csv", index_col=0)

        ## Define the function which determines whether a state is closer to controls/patients
        dfdata = states.loc[[i for i in states.index if (i != "annotation")]]
        samples = states.loc["annotation"]
        return dfdata, samples, seed_number, SIMU_params, states, patients, targets, file_folder

    def test_compute_frontier(self):
        from NORDic.NORDic_DS.functions import compute_frontier
        dfdata, samples, seed_number, SIMU_params, states, patients, targets, file_folder = self.create_instance()
        ## compute_frontier fits a SVM model to the data
        frontier = compute_frontier(dfdata, samples)
        score = lambda attrs : (frontier.predict(attrs.values.T)==1).astype(int)
        return score, seed_number, SIMU_params, states, patients, targets, file_folder

    def test_drug_repurposing(self):
        score, seed_number, SIMU_params, states, patients, targets, file_folder = self.test_compute_frontier()
        ## Drug Repurposing through adaptive testing
        from NORDic.NORDic_DR.functions import adaptive_testing
        BANDIT_args = {
            'bandit': 'LinGapE', 
            'seed': seed_number,
            'delta': 0.1, 
            'nsimu': 1, 
            'm': 4, 
            'c': 0, 
            'sigma': 1,
            'beta': "heuristic",
            'epsilon': 0.1,
            'tracking_type': "D",
            'gain_type': "empirical",
            'learner': "AdaHedge"
        }
        recommendation = adaptive_testing(file_folder+"solution.bnet", targets, targets, score, patients, 
                SIMU_params, BANDIT_args, reward_fname=None, quiet=False).T
        self.assertEqual(recommendation.shape[0],targets.shape[1])
        self.assertTrue(np.isclose(recommendation.sum().sum(), BANDIT_args["m"]))

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import numpy as np
import pandas as pd
import json
from subprocess import check_output as sbcheck_output
from time import sleep
import os
from glob import glob
import io
import requests
import pickle

from NORDic.UTILS.STRING_utils import get_protein_names_from_STRING
from NORDic.UTILS.LINCS_utils import build_url, post_request, lincs_api_url

#https://biodbnet-abcc.ncifcrf.gov/webServices/RestWebService.php

def request_biodbnet(probe_list, from_, to_, taxon_id, chunksize=500, quiet=False):
    '''
    Converts gene identifier from from to to in a given species

    ...

    Parameters
    ----------
    probe_list : Python character string list
        list of probes to convert (of type from_)
    from_ : Python character string
        an identifier type as recognized by BioDBnet
    to_ : Python character string
        an identifier type as recognized by BioDBnet
    taxonId : Python integer
        NCBI taxonomy ID
    chunksize : Python integer
        [default=500] : 1 chunk per request

    Returns
    ----------
    res_df : Pandas DataFrame
        rows/["InputValue"/from_] x columns[to_] ("-" if the identifier has not been found)
    '''
    chunk_probes=[probe_list[i:i+chunksize] for i in range(0,len(probe_list),chunksize)]
    res_list = []
    biodbnet_url="https://biodbnet-abcc.ncifcrf.gov/webServices/rest.php/biodbnetRestApi.json"
    for i, chunk in enumerate(chunk_probes):
        if (not quiet):
            print("<BioDBNet> %d/%d" % (i+1, len(chunk_probes)))
        args_di = {
            "method":"db2db",
            "format": "row",
            "inputValues":",".join(chunk),
            "input":from_,
            "outputs":to_,
            "taxonId":str(taxon_id),
        }
        #https://biodbnet-abcc.ncifcrf.gov/webServices/RestSampleCode.php
        query=biodbnet_url+"?"+"&".join([k+"="+args_di[k] for k in args_di])
        res_list += json.loads("; ".join(sbcheck_output("wget -qO- \""+query+"\"", shell=True).decode("utf-8").split("//")))
        sleep(1)
    res_df = pd.DataFrame(res_list)
    res_df.index = res_df["InputValue"]
    res_df = res_df.drop(columns=["InputValue"])
    return res_df

EntrezGene_missing_genes = {
    "ENSP00000451560": "TPPP2",
    "C11ORF74": "IFTAP",
    "RP11-566K11.2": "TUBB4",
}
def convert_genes_EntrezGene(gene_list, taxon_id, app_name, chunksize=100, missing_genes=EntrezGene_missing_genes,quiet=False):
    '''
    Convert gene symbols into EntrezGene CID

    ...

    Parameters
    ----------
    gene_list : Python character string list
        list of genes
    taxon_id : Python character string
        NCBI taxonomy ID
    app_name : Python character string
        STRING identifier
    missing_genes : Python dictionary of character string x character string
        known conversions
    chunksize : Python integer
        [default=100] : 1 chunk per request
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    res_df : Pandas DataFrame
        rows/["InputValue"] x columns/["Gene ID"/might be separated by "; "] ("-" if they do not exist) or None if no identifier has been found
    '''
    if (not quiet):
        print("<UTILS_DATA> Gene Symbol -> Gene ID (%d probes)" % len(gene_list))
    probes = request_biodbnet(gene_list, "Gene Symbol and Synonyms", "Gene ID", taxon_id, chunksize=chunksize)
    other_ids = list(probes[probes["Gene ID"]=="-"].index)
    probes = probes[probes["Gene ID"]!="-"]
    df_list = [probes]
    if (len(other_ids)>0):
        if (not quiet):
            print("<UTILS_DATA> Ensembl Gene ID -> Gene ID (%d probes)" % len(other_ids))
        other_probes = request_biodbnet(other_ids, "Ensembl Gene ID", "Gene ID", taxon_id, chunksize=chunksize)
        other_ids = list(other_probes[other_probes["Gene ID"]=="-"].index)
        other_probes = other_probes[other_probes["Gene ID"]!="-"]
        df_list.append(other_probes)
    if (len(other_ids)>0):
        if (not quiet):
            print("<UTILS_DATA> HGNC ID -> Gene ID (%d probes)" % len(other_ids))
        other_other_probes = request_biodbnet(other_ids, "HGNC ID", "Gene ID", taxon_id, chunksize=chunksize)
        other_ids = list(other_other_probes[other_other_probes["Gene ID"]=="-"].index)
        other_other_probes = other_other_probes[other_other_probes["Gene ID"]!="-"]
        df_list.append(other_other_probes)
    if (len(other_ids)>0 and app_name is not None):
        if (not quiet):
            print("<UTILS_DATA> STRING ID -> Gene ID (%d probes)" % len(other_ids))
        res_df = get_protein_names_from_STRING(other_ids, taxon_id, app_name)
        ## if can't be found automatically...
        if (res_df is not None):
            other_ids = []
            for idx in list(res_df["queryItem"]):
                other_ids.append(missing_genes.get(idx, idx))
            res_df = request_biodbnet(other_ids, "Gene Symbol and Synonyms", "Gene ID", taxon_id, chunksize=chunksize, quiet=quiet)
            df_list.append(res_df)
    if (len(other_ids)>0):
        df_list.append(pd.DataFrame([], index=other_ids, columns=["Gene ID"]).fillna("-"))
    if (len(df_list)>0):
        probes = pd.concat(tuple(df_list), axis=0)
    else:
        probes = None
        return probes
    probes = probes.sort_index(ascending=True)
    if (not quiet):
        print("%d probes (successful %d, unsuccessful %d)" % (len(probes), len(probes.loc[probes["Gene ID"]!='-']), len(probes.loc[probes["Gene ID"]=="-"])))
    return probes

def convert_EntrezGene_LINCSL1000(file_folder, EntrezGenes, user_key, quiet=False):
    '''
    Converts EntrezIDs to Gene Symbols present in LINCS L1000

    ...

    Parameters
    ----------
    file_folder : Python character string
        path to folder of intermediate results
    EntrezGenes : Python character string list
        list of EntrezGene IDs
    user_key : Python character string
        LINCS L1000 user key
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    Pandas : Pandas DataFrame
        rows/[EntrezID] x columns/["Gene Symbol","Entrez ID"] ("-" if they do not exist)
    '''
    assert user_key
    gene_file = file_folder+"entrezGene_LINCSL1000.pck"
    if (not os.path.exists(gene_file)):
        pert_inames = [None]*len(EntrezGenes)
        entrez_ids = [None]*len(EntrezGenes)
        seen_genes = []
    else:
        with open(gene_file, "rb") as f:
            results = pickle.load(f)
        pert_inames, entrez_ids, seen_genes = results["pert_inames"], results["entrez_ids"], results["seen_genes"]
    for ig, g in enumerate(EntrezGenes):
        if (g in seen_genes):
                continue
        if (not quiet):
            print("<UTILS_DATA> Entrez ID -> LINCS L1000 (%d/%d)" % (ig+1,len(EntrezGenes)))
        endpoint="genes"
        method="filter"
        all_entrezid = str(g).split("; ")
        for entrezid in all_entrezid:
            params = {"where":{"entrez_id": str(entrezid)},"fields":["gene_symbol"]}
            request_url = build_url(endpoint, method, params, user_key=user_key)
            data = post_request(request_url, quiet=True, pause_time=0.3)
            if (len(data)==0):
                continue
            else:
                pert_inames[ig] = data[0]["gene_symbol"]
                entrez_ids[ig] = entrezid
                if (pert_inames[ig]==g):
                    break
        if (pert_inames[ig] is not None):
            print("\t".join([g, pert_inames[ig], str(ig+1), str(len(EntrezGenes))]))
        seen_genes.append(g)
        with open(gene_file, "wb") as f:
            pickle.dump({"pert_inames":pert_inames, "entrez_ids": entrez_ids, "seen_genes": seen_genes}, f)
    pert_inames_ = [p if (p is not None) else "-" for p in pert_inames]
    entrez_ids_ = [entrez_ids[i] if (pert_inames[i] is not None) else "-" for i in range(len(EntrezGenes))]
    pert_df = pd.DataFrame([pert_inames_, entrez_ids], columns=EntrezGenes, index=["Gene Symbol", "Entrez ID"]).T
    pert_df = pert_df.sort_values(by="Gene Symbol", ascending=True)
    pert_df = pert_df.loc[~pert_df.duplicated()]
    return pert_df

def get_all_celllines(pert_inames, user_key, quiet=False):
    '''
    Get all cell lines in which one gene in the input list has been specifically perturbed (genetic perturbation)

    ...

    Parameters
    ----------
    pert_inames : Python character string
        List of genes (symbols from LINCS L1000)
    user_key : Python character string
        user key from LINCS L1000 CLUE API
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    cell_lines : Python character string list
        list of cell lines in which at least one gene from pert_inames has been perturbed
    '''
    assert user_key
    endpoint = "sigs"
    method = "filter"
    cell_lines = []
    for ig, g in enumerate(pert_inames):
        if (not quiet):
            print("<UTILS_DATA> Perturbagen %s (%d/%d)" % (g, ig+1, len(pert_inames)), end=": ")
        params = {"where": {"pert_iname": g}, "fields": ["cell_id"]}
        request_url = build_url(endpoint, method, params, user_key=user_key)
        response = requests.get(request_url)
        assert response.status_code == 200
        data = json.loads(response.text)
        cell_lines_gene = list(set([d["cell_id"] for d in data]))
        if (not quiet):
            print("%d unique cells" % len(cell_lines_gene))
        cell_lines = list(set(cell_lines_gene+cell_lines))
    return cell_lines



# coding: utf-8

import numpy as np
import pandas as pd
from sklearn.metrics import pairwise_distances
from qnorm import quantile_normalize as qnm

def quantile_normalize(df, njobs=1):
    return qnm(df, axis=1, ncpus=njobs)

def binarize_experiments(data, thres=0.5, method="binary", strict=True, njobs=1):
    '''
    Binarize experimental profiles

    ...

    Parameters
    ----------
    data : Pandas DataFrame
        rows/[genes] x columns/[samples]
    thres : Python float
        [default=0.5] : threshold for @method="binary" (in [0,0.5])
    method : Python character string
        [default="binary"] : binarization method in ["binary","probin"]
    strict : Python bool
        [default=True] : takes into account equalities (if set to True, value=thres will lead to undefined for the corresponding gene)
    njobs : Python integer
        [default=1] : parallelism if needed

    Returns
    ----------
    signatures : Pandas DataFrame: rows/[genes] x columns[samples] with values in [0,1,NaN]
    '''
    assert method in ["binary","probin"]
    assert thres <= 0.5 and thres >= 0
    from multiprocessing import cpu_count
    assert njobs <= max(1,cpu_count()-2) and njobs >= 1
    if (method == "probin"):
        ## /!\ needs to install profilebinR (see install_env_synthesis.sh)
        from profile_binr import ProfileBin
        probin = ProfileBin(data.T)
        probin.fit(njobs)
        data = probin.binarize().T
        return data
    else:
        assert thres <= 0.5 and thres >= 0
        signatures = quantile_normalize(data, njobs=njobs)
        min_mat = pd.concat([signatures.min(axis=1)] * signatures.shape[1], axis=1, ignore_index=True).values
        sigs = pd.DataFrame(signatures.values-min_mat, index=signatures.index, columns=signatures.columns)
        max_mat = pd.concat([sigs.max(axis=1)] * sigs.shape[1], axis=1, ignore_index=True).values
        signatures = pd.DataFrame(sigs.values/max_mat, index=signatures.index, columns=signatures.columns)
        if (strict):
            signatures[signatures < thres] = 0
            signatures[signatures > 1-thres] = 1
            signatures[(signatures>=thres)&(signatures<=1-thres)] = np.nan
        else:
            signatures[signatures<=thres] = 0
            signatures[signatures>=1-thres] = 1
            signatures[(signatures>thres)&(signatures<1-thres)] = np.nan
        return signatures

def compare_states(x, y, genes=None):
    '''
    Computes the similarity between two sets of Boolean states

    ...

    Parameters
    ----------
    x : Pandas DataFrame
        rows/[genes] x columns/[state IDs] contains (0, 1, NaN)
    y : Pandas DataFrame
        rows/[genes] x columns/[state IDs] contains (0, 1, NaN)
    genes : Python character string list
        list of gene symbols 

    Returns
    ----------
    sims : NumPy array
        similarities between each column of x and each columns of y, on the list of N present genes in genes (if provided) otherwise on the union of N genes in x and y
    N : Python integer
        number of genes on which the similarity is computed
    '''
    assert all([(zx in [0,1] or np.isnan(zx)) for zx in np.unique(x.values)])
    assert all([(zy in [0,1] or np.isnan(zy)) for zy in np.unique(y.values)])
    xx = pd.DataFrame(x.values, index=x.index, columns=["X%d" %d for d in range(x.shape[1])])
    yy = pd.DataFrame(y.values, index=y.index, columns=["Y%d" %d for d in range(y.shape[1])])
    z = xx.join(yy, how="outer")
    if (genes is not None):
        gene_list = [g for g in genes if (g in z.index)]
        z = z.loc[gene_list]
        if (len(gene_list)==0):
            raise ValueError("None of the genes in the list is present in any of the vectors!")
    N = z.shape[0]
    x_, y_ = [z[u.columns] for u in [xx,yy]]
    ## Compute separately distances between 1's and 0's
    X_pos, Y_pos, X_neg, Y_neg = [np.mod(u.fillna(v).T.values+1, 2).astype(float) for v in [0,1] for u in [x_,y_]]
    dists_pos, dists_neg = [pairwise_distances(X, Y, metric="cityblock")/N for X,Y in [[X_pos,Y_pos],[X_neg,Y_neg]]]
    dists_pos[np.isnan(dists_pos)] = 0
    dists_neg[np.isnan(dists_neg)] = 0
    dists = np.power(np.multiply(dists_pos,dists_neg), 0.5) # geometric mean
    ## Numerical approximations
    dists[np.isclose(dists, 0)] = 0
    dists[np.isclose(dists, 1)] = 1
    sims = 1-dists
    return sims, N

def finetune_binthres(df, samples, network_fname, mutation, step=0.005, maxt=0.5, mint=0, score_binthres=lambda itc,ita_c,ita_t:(1-itc)*ita_c*ita_t, njobs=1, verbose=True):
    '''
    Select the binarization threshold (in function @binarize_experiments) which maximize the dissimilarity interconditions and the similarity intracondition
    ...

    Parameters
    ----------
    df : Pandas DataFrame
        rows/[genes] x columns/[samples]: profiles
    samples : Python character string list
        annotations of conditions for each sample in df
    network_fname : Python character string
        file name containing the network
    mutation : Python dictionary
        dictionary (key=gene, value=perturbation type) gene perturbations which are considered
    step : Python float
        [default=0.005] step in the interval to look for the threshold value
    maxt : Python float
        [default=0.5] maximum threshold value
    mint : Python float
        [default=0.] minimum threshold value
    score_binthres : Python lambda function
        [default=lambda itc,ita_c,ita_t:(1-itc)*ita_c*ita_t] fitness function for the threshold value
    njobs : Python integer
        [default=1] number of parallel jobs
    verbose : Python bool
        [default=True] prints out verbose

    Returns
    ----------
    max_thres : Python float
        threshold value maximizing the fitness function
        
    '''
    assert mint>=0 and maxt<=0.5
    with open(network_fname, "r") as f:
        gene_list = [x.split(", ")[0] for x in f.read().split("\n") if (len(x)>0)]
    gene_outputs = [g for g in gene_list if (g not in mutation)]
    max_thres, max_score = None, -float("inf")
    scale=float("0."+("0"*(len(str(step).split(".")[-1])-1))+"1")
    for bt in [x*scale for x in range(int(mint/scale),int(maxt/scale)+1,int(step/scale))]:
        if (bt>maxt):
            break
        df_b = binarize_experiments(df, thres=bt, method="binary", strict=True, njobs=njobs)
        cp = df_b[[s for s in samples if (samples[s]=="control")]]
        tps = df_b[[s for s in samples if (samples[s]=="treated")]]
        sims_interconditions, _ = compare_states(cp, tps, gene_outputs)
        sims_intractrl, _ = compare_states(cp, cp, gene_outputs)
        sims_intratps, _ = compare_states(tps, tps, gene_outputs)
        sc = score_binthres(np.max(sims_interconditions), np.min(sims_intractrl), np.min(sims_intratps))
        #sc = score_binthres(np.max(sims_interconditions), np.max(sims_intractrl), np.max(sims_intratps))
        if (sc > max_score):
            max_thres, max_score = bt, sc
        if (verbose):
            print("BinThres=%.3f\tMax.sim CTRL||TRT: %.3f\tMin.sim CTRL||CTRL: %.3f\tMin.sim TRT||TRT: %.3f\tScore: %.3f"%(bt, np.max(sims_interconditions), np.min(sims_intractrl), np.min(sims_intratps), sc))
    if (verbose):
        print("Final bin_thres value: %.3f (score %.3f)" % (max_thres, max_score))
    return max_thres

## Tests
if __name__ == "__main__":
    ## Length of vectors
    N=10
    P=2
    ## All 1
    ones = pd.DataFrame([1]*N, index=range(N))
    ## Half 1, Half 0
    half = pd.DataFrame([1]*(N//P)+[0]*(N-N//P), index=range(N))
    ## Half 1, Half NaN
    halfnan = pd.DataFrame([1]*(N//P)+[np.nan]*(N-N//P), index=range(N))
    ## Half 0, Half NaN
    nanhalf = pd.DataFrame([0]*(N//P)+[np.nan]*(N-N//P), index=range(N))
    ## All 0
    zeros = pd.DataFrame([0]*N, index=range(N))
    tests = [ones, half, halfnan, nanhalf, zeros]
    res_mat = np.zeros((len(tests), len(tests)))
    all_pairs = [[tests[i], tests[j]] for i in range(len(tests)) for j in range(len(tests))]
    idx_lst = ["%s" % (list(u[u.columns[0]].value_counts(dropna=False).index.astype(float))) for u in tests]
    res_mat = np.array([float(compare_states(c1, c2)[0]) for c1, c2 in all_pairs]).reshape((len(tests), len(tests)))
    res_df = pd.DataFrame(res_mat, index=idx_lst, columns=idx_lst)
    print(res_df.T)



#coding: utf-8

import NORDic
import os
import pandas as pd
from multiprocessing import cpu_count
import numpy as np
import mpbn
import unittest

## II. Prioritization of master regulators (**NORDic PMR**)

class TestPMR(unittest.TestCase):

    def create_instance(self):
        ## Path to files
        seed_number=123456
        njobs=min(5,max(1,cpu_count()-2))
        file_folder="ToyOndine/"
        ## Options for PMR
        np.random.seed(seed_number)
        with open(file_folder+"solution.bnet", "r") as f:
            genes = [line.split(", ")[0] for line in f.read().split("\n") if (len(line)>0)]
        k=1
        IM_params = {"seed": seed_number, "gene_inputs": genes, "gene_outputs": genes}
        SIMU_params = {'nb_sims': 1000, 'rates': "fully_asynchronous", 'thread_count': njobs, 'depth': "constant_unitary"}
        ## Generate "patient" states
        ### Create the mutated networks
        f_wildtype = mpbn.load(file_folder+"solution.bnet")
        f_mutant = mpbn.MPBooleanNetwork(f_wildtype)
        f_mutant["PHOX2B"] = 0 ## CCHS ~ PHOX2B mutant
        ### Initial states drawn at random
        state_len = 20
        states = pd.DataFrame(
              [np.random.choice([0,1], p=[0.5,0.5], size=len(genes)).tolist() for _ in range(state_len)]
              , columns=genes, index=range(state_len)).T
        A_WT, A_mut = [], []
        for state in states.columns:
            x0 = f_wildtype.zero()
            for i in list(states.loc[states[state]==1].columns):
                x0[i] = 1
            ### Get the reachable attractors from initial state in the presence/absence of X0 knockout
            A_WT += [a for a in list(f_wildtype.attractors(reachable_from=x0))]
            A_mut += [a for a in list(f_mutant.attractors(reachable_from=x0))]
        for a in A_WT:
            a.update({"annotation": 1})
        for a in A_mut:
            a.update({"annotation": 2})
        controls = {"Control%d" % (i+1): a for i,a in enumerate(A_WT)}
        patients = {"Case%d" % (i+1): a for i,a in enumerate(A_mut)}
        controls.update(patients)
        states = pd.DataFrame(controls)
        states.to_csv(file_folder+"phenotypes.csv")
        patients = pd.DataFrame(patients)
        patients = patients.loc[[i for i in patients.index if (i != "annotation")]]
        return file_folder, k, patients, IM_params, SIMU_params

    def test_prioritization_master_regulators(self):
        file_folder, k, patients, IM_params, SIMU_params = self.create_instance()
        ## Prioritization of Master Regulators
        from NORDic.NORDic_PMR.functions import greedy
        S, spreads = greedy(file_folder+"solution.bnet", k, patients, IM_params, SIMU_params, save_folder=file_folder)
        self.assertEqual(spreads.shape[0],6)
        self.assertEqual(np.max(spreads.values),spreads.loc[S[0][0]][str(S)])

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import pandas as pd
import numpy as np
import random
from tqdm import tqdm
from scipy.stats.mstats import gmean
from subprocess import call as sbcall
from multiprocessing import cpu_count
from time import time

from mpbn import load, MPBooleanNetwork
from mpbn.simulation import MPSimMemory, step, constant_maximum_depth, uniform_rates, nexponential_depth, nexponential_rates, constant_unitary_depth, fully_asynchronous_rates
from bonesis import BoNesis
import maboss

import matplotlib.pyplot as plt
import matplotlib.ticker as mticker

from NORDic.UTILS.utils_state import compare_states, binarize_experiments
from NORDic.UTILS.utils_grn import solution2influences, create_grn

import mpbn
import mpbn.simulation as mpbn_sim
from time import time
from sklearn.metrics.pairwise import nan_euclidean_distances as ndist

from numpy.random import choice
from random import seed as rseed
from numpy.random import seed as npseed
from mpbn.simulation import MPBNSim
from pandas import DataFrame

from joblib import Parallel, delayed, wrap_non_picklable_objects
from joblib.externals.loky import set_loky_pickler

import contextlib
@contextlib.contextmanager
def capture():
    import sys
    from io import StringIO
    oldout,olderr = sys.stdout, sys.stderr
    try:
        out=[StringIO(), StringIO()]
        sys.stdout,sys.stderr = out
        yield out
    finally:
        sys.stdout,sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()

def test(enumerator, seednb, njobs, network_fname, control_profile, treated_profiles, compare_to, mutation_permanent={}, mutation_transient={}, gene_outputs=None, print_boxplot=False, verbose=True):
    assert compare_to in ["profiles","attractors"]
    sims_di = {}
    for c in control_profile.columns:
        sims_c = {}
        for t in treated_profiles.columns:
            net_c = enumerator(seednb,njobs)
            attrs1 = net_c.up_to_attractors(network_fname, control_profile[[c]], treated_profiles[[t]], mutation_permanent, mutation_transient, verbose)
            attrs1.columns = ["Attr%d_%s_1" % (i,c) for i in range(attrs1.shape[1])]
            if (compare_to=="profiles"):
                attrs2 = treated_profiles[[t]]
            else:
                net_t = enumerator(seednb,njobs)
                attrs2 = net_t.up_to_attractors(network_fname, treated_profiles[[t]], treated_profiles[[t]], {}, {}, verbose)
                attrs2.columns = ["Attr%d_%s_2" % (i,t) for i in range(attrs2.shape[1])]
            sims = net_c.attrs_similarity(attrs1, attrs2, gene_outputs=gene_outputs)
            sims_c.setdefault(t, sims)
            if(verbose):
                if (np.isnan(attrs1).values.all()&np.isnan(attrs2).values.all()):
                    print("NaN (not reachable)")
                    print("Similarity %s" % str(sims))
                    continue
                elif (np.isnan(attrs1).values.all()):
                    print("Treated %s (showing 10 first out of %d, Control %s: NaN, 5/%d genes)" % (t, attrs2.shape[1], c, attrs2.shape[0]))
                    #print(attrs2.iloc[:,:10].dropna(how="all").head())
                    if (any([g in gene_outputs for g in mutation])):
                        for g in [g for g in mutation if (g in gene_outputs)]:
                            print("".join(list(attrs2.loc[g].iloc[:,:10].astype(str))))
                    print("Similarity %s" % str(sims))
                    continue
                elif (np.isnan(attrs2).values.all()):
                    print("Control %s (showing 10 first out of %d, Treated %s: NaN, 5/%d genes)" % (c, attrs1.shape[1], t, attrs2.shape[0]))
                    #print(attrs1.iloc[:,:10].dropna(how="all").head())
                    if (any([g in gene_outputs for g in mutation])):
                        for g in [g for g in mutation if (g in gene_outputs)]:
                            print("".join(list(attrs1.loc[g].iloc[:,:10].astype(str))))
                    print("Similarity %s" % str(sims))
                    continue   
                A_set = attrs1.iloc[:,:10].join(attrs2.iloc[:,:10], how="outer").dropna(how="any")
                A_set = A_set.loc[[g for g in gene_outputs if (g in A_set.index)]]         
                print("Control %s, Treated %s (showing 10 first out of %d, resp. 10/%d, %d genes)" % (c,t,attrs1.shape[1],attrs2.shape[1], A_set.shape[0]))
                subst = (A_set-A_set[attrs2.columns[0]]).mean(axis=1)
                if (A_set.shape[1]<=2):
                    print(A_set)
                else:
                    if ((subst!=0).any()):
                        print(A_set.loc[(subst!=0)])
                    else:
                        print("Attractors are equal.")
                mutation = list(set([g for g in mutation_permanent]+[g for g in mutation_transient]))
                if (any([g in A_set.index for g in mutation]) and any([subst.loc[g]==0 for g in mutation])):
                    print(A_set.loc[(subst==0)].loc[[g for g in mutation if (g in A_set.index)]])
                print("Similarity %s" % str(sims))
        sims_di.setdefault(c, sims_c)
    net_c.sims = sims_di
    if (print_boxplot):
        net_c.boxplot()
    net_c.max_values = {"%s->%s"%(c,t):np.max(net_c.sims[c][t]) for c in net_c.sims for t in net_c.sims[c]}
    #gmean = lambda ls: np.prod(np.power(ls,1/len(ls)))
    gmeansim = gmean([net_c.max_values[m]+1 for m in net_c.max_values])-1
    return gmeansim, net_c

## sims = {control: {treated: [[sim]] (or [[sim(a)] for all attractors])}
class BN_SIM(object):
    def __init__(self, seednb=0,njobs=None):
        self.seednb = seednb
        self.initial = None
        self.mutation_permanent = {}
        self.mutation_transient = {}
        self.all_mutants = []
        self.gene_outputs = None
        self.sims = None
        self.attrs = None
        self.network = None
        self.njobs = njobs if (njobs is not None) else max(1,cpu_count()-2)

    def boxplot(self):
        assert self.sims is not None
        f_lst = []
        for im, mut_di in enumerate([self.mutation_transient, self.mutation_permanent]):
            f_lst += [g+(":=" if (im==1) else "~")+("OE" if (gval>0) else "KO") for g, gval in mut_di.items()]
        f = "_".join(f_lst)
        controls = list(self.sims.keys())
        treated = list(self.sims[controls[0]].keys())
        ncontrol, ntreated = len(controls), len(treated)
        fig, axes = plt.subplots(nrows=ntreated, ncols=ncontrol, figsize=(5*(1.5**int(ncontrol>1)),5*(1.5**int(ntreated>1))))
        for ic, c in enumerate(controls):
            for it, t in enumerate(treated):
                if (ntreated==1):
                    ax = axes[ic]
                elif (ncontrol==1):
                    ax = axes[it]
                else:
                    ax = axes[it][ic]
                ticks_loc = ax.get_xticks().tolist()
                ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
                ax.set_xticklabels([f for x in ticks_loc])
                ax.set_ylabel("%s\n(N=%d)||%s(N=%d)" % (c, self.sims[c][t].shape[0], t, self.sims[c][t].shape[1]))
                ax.boxplot(self.sims[c][t].flatten())
                avg = np.mean(self.sims[c][t])
                ax.text(1.1, avg, "mean=%.2f" % avg, color="black", size=10)
                ax.set_title("%s\n%s vs %s" % (f, c, t))
        plt.show()

    def attrs_similarity(self, attrs1, attrs2, gene_outputs=None):
        assert self.network is not None
        if (np.isnan(attrs1.values).all() or np.isnan(attrs2.values).all()):
            return np.array([[0.]])
        if (gene_outputs is None):
            self.gene_outputs = [g for g in self.gene_list if (g not in [m for m in self.mutation_permanent]+[m for m in self.mutation_transient])]
        else:
            self.gene_outputs = gene_outputs
        sims, ngenes = compare_states(attrs1, attrs2, self.gene_outputs)
        if (ngenes!=len(self.gene_outputs)):
            print("/!\ %d/%d genes (initially %d %d)" % (ngenes, len(self.gene_outputs), attrs1.shape[0], attrs2.shape[0]))
        return sims

    def update_network(self, network_fname, initial, final=None, mutation_permanent={}, mutation_transient={}, verbose=True):
        if (final is None):
            final = initial
        assert initial.shape[1]==1 and final.shape[1]==1
        self.initialize_network(network_fname)
        if (verbose):
            print("Initialized network")
        with open(network_fname, "r") as f:
            grf_list = f.read().split("\n")
        grfs = dict([x.split(", ") for x in grf_list if (len(x)>0)])
        input_genes = [g for g in grfs if (grfs[g] in ["0","1"])] ## genes with no regulators
        self.mutation_permanent = {inp: int(initial.loc[inp][initial.columns[0]]) for inp in input_genes if (inp in initial.dropna().index)}
        N = len(self.mutation_permanent)
        self.mutation_permanent.update(mutation_permanent)
        if (verbose):
            print("Initialized %d permanent mutations (%d from initial state)" % (len(self.mutation_permanent), N))
        self.mutation_transient.update(mutation_transient)
        if (verbose):
            print("Initialized %d transient mutations" % len(self.mutation_transient))
        self.all_mutants = list(set([g for g in self.mutation_permanent]+[g for g in self.mutation_transient]))
        if (verbose):
            print("... Total of %d (unique) mutations" % len(self.all_mutants))
        self.add_transient_mutation(self.mutation_transient)
        self.add_permanent_mutation(self.mutation_permanent)
        self.add_initial_states(initial, final)
        if (verbose):
            print("Initialized initial state")
        return None

    def up_to_attractors(self, network_fname, initial, final, mutation_permanent={}, mutation_transient={}, verbose=True):
        self.update_network(network_fname, initial, final, mutation_permanent, mutation_transient, verbose=verbose)
        attrs = self.enumerate_attractors(verbose)
        if (verbose):
            print("Enumerated %d attractors" % attrs.shape[1])
        return attrs

    def initialize_network(self, network_fname):
        raise NotImplemented

    def add_initial_states(self, initial, final=None):
        raise NotImplemented

    def add_transient_mutation(self, mutation):
        raise NotImplemented

    def add_permanent_mutation(self, mutation):
        raise NotImplemented

    def enumerate_attractors(self, verbose=False):
        raise NotImplemented

    def generate_trajectories(self, params={}, outputs=[]):
        raise NotImplemented

############ MPBN VERSION

class MPBN_SIM(BN_SIM):
    def __init__(self, seednb=0,njobs=None):
        super(MPBN_SIM, self).__init__(seednb,njobs)
        self.attrs = []

    def initialize_network(self, network_fname):
        self.network = MPBooleanNetwork(load(network_fname))
        self.gene_list = [g for g in self.network.zero()]

    def add_initial_states(self, initial, final=None):
        assert initial.shape[1]==1
        x0 = self.network.zero()
        for i in initial.loc[~pd.isnull(initial[initial.columns[0]])].index:
            x0[i] = int(initial.loc[i][initial.columns[0]])
        for g, gval in self.mutation_transient.items():
            x0[g] = int(gval)
        self.initial = x0

    def add_transient_mutation(self, mutation):
        pass

    def add_permanent_mutation(self, mutation):
        for i, fi in mutation.items():
            self.network[i] = fi

    def enumerate_attractors(self, max_attrs=-1, verbose=True):
        random.seed(self.seednb)
        np.random.seed(self.seednb)
        #attrs_gen = self.network.fixedpoints(reachable_from=self.initial)
        if (verbose):
            attrs_gen = self.network.attractors(reachable_from=self.initial)
        else:
            with capture() as out:
                attrs_gen = self.network.attractors(reachable_from=self.initial)
        if (max_attrs>0):
            attrs = [None]*max_attrs
            idt = lambda x : x
            for ia, a in (tqdm if (verbose) else idt)(enumerate(attrs_gen)):
                self.attrs.append(a)
                attrs[ia] = pd.DataFrame({"Attr%d"%ia: a})
                if (max_attrs>0 and (ia==max_attrs-1)):
                    break
        else:
            self.attrs = [a for a in (tqdm if (verbose) else idt)(attrs_gen)]
            attrs = [pd.DataFrame({"Attr%d"%ia: a}) for ia, a in enumerate(self.attrs)]
        attrs_df = attrs[0].join(attrs[1:], how="outer")
        self.attrs = attrs_df
        return attrs_df

    def generate_trajectories(self, params={}, outputs=[], show_plot=True):
        rseed(self.seednb)
        npseed(self.seednb)
        nsims = params.get('sample_count', 1000)
        njobs = params.get('thread_count', 1)#self.njobs)
        max_time = params.get("max_time", 1000)
        noutputs = [g for g in (self.gene_list if (len(outputs)==0) else outputs) if (g not in self.all_mutants)]
        seeds = choice(range(int(max(1e4,nsims))), size=nsims)
        name_state = lambda s : " -- ".join(list(sorted([g for g in s if ((s[g]==1) and (g in noutputs))]))) if (any([(s[g]==1) and (g in noutputs) for g in s])) else "<nil>"
        #@delayed
        #@wrap_non_picklable_objects
        def generate_trajectory(snb, net, params, istates, noutputs):
            rseed(snb)
            npseed(snb)
            netsim = MPBNSim(net)
            depth = params.get("depth", constant_maximum_depth)(netsim)
            W = params.get("W", uniform_rates)(netsim)
            n, k, x = len(netsim), 0, istates.copy()
            mem = MPSimMemory(n, n)
            str_state = name_state(x)
            trajectory = [] # [k: state at k]
            trajectory.append([0,str_state,1])
            while (k<=max_time):
                k += 1
                r = step(netsim, mem, x, depth, W)
                str_state = name_state(x)
                trajectory.append([k,str_state,1])
                if (not r):
                    for ki in range(k+1, max_time):
                        trajectory.append([ki,str_state,1])
                    break
            return trajectory
        def plot_trajectory(probs):
            plt.figure(figsize=(8,5))
            plt.xlim((-max_time//20,max_time+max_time//20))
            plt.ylim((-0.1,1.1))
            for c in probs.columns:
                plt.plot(range(max_time), list(probs[c]), label=c)
            plt.xlabel("")
            plt.ylabel("")
            plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
            plt.show()
        if ((njobs==1) or (nsims==1)):
            trajectories = [y for s in seeds for y in generate_trajectory(s, self.network, params, self.initial, noutputs)]
        else:
            set_loky_pickler()
            parallel = Parallel(n_jobs=njobs, backend='loky')
            trajectories = [y for x in parallel(delayed(generate_trajectory)(s, self.network, params, self.initial, noutputs) for s in seeds) for y in x]
        trajectories = pd.DataFrame(trajectories, index=range(nsims*max_time), columns=["step","state","count"])
        probs = pd.pivot_table(trajectories, values="count", index="step", columns="state", aggfunc='count').fillna(0)/nsims
        if (show_plot):
            plot_trajectory(probs)
        table = pd.DataFrame([list(probs.loc[max_time-1])], index=["prob"], columns=["{"+",".join([g+"=1" for g in c.split(" -- ")])+"}" if (c!="<nil>") else c for c in probs.columns])
        table = table[[c for c in table.columns if (float(table.loc["prob"][c])!=0)]]
        return table

############ BONESIS VERSION 

class BONESIS_SIM(BN_SIM):
    def __init__(self, seednb=0,njobs=None):
        super(BONESIS_SIM, self).__init__(seednb,njobs)
        self.nattr = 0

    def initialize_network(self, network_fname):
        with open(network_fname, "r") as f:
            solution = pd.Series(dict([x.split(", ") for x in f.read().split("\n")]))
        influences = solution2influences(solution)
        influences.index = ["_".join(x.split("-")) for x in influences.index]
        influences.columns = ["_".join(x.split("-")) for x in influences.columns]
        self.gene_list = list(set(list(influences.index)+list(influences.columns)))
        self.grn = create_grn(influences, exact=True, quiet=True)
        #self.network = list(BoNesis(self.grn, {}).boolean_networks())[0]

    def add_initial_states(self, initial, final):
        initial_ = pd.DataFrame(initial.values, index=initial.index, columns=["initial"]).dropna().astype(int)
        initial_ = initial_.T
        for g, gval in self.mutation_transient.items():
            initial_[g] = int(gval)
        initial_ = initial_.T
        final_ = pd.DataFrame(final.values, index=final.index, columns=["final"])
        data_df = final_.join(initial_, how="outer")
        data_df.columns = ["exp","init"]
        data_df.index = ["_".join(x.split("-")) for x in data_df.index]
        data_exps = data_df[["exp"]].dropna().to_dict()
        data_exps.update(data_df[["init"]].dropna().to_dict())
        self.network = BoNesis(self.grn, data_exps)
        self.initial = initial_.to_dict()["initial"]
        #self.network = list(BoNesis(self.grn, data_exps).boolean_networks())[0]

    def add_transient_mutation(self, mutation):
        pass

    def add_permanent_mutation(self, mutation):
        pass

    def enumerate_attractors(self, verbose=True):
        #attrs = pd.DataFrame({"Attr%d"%i : attr for i, attr in enumerate(self.network.attractors(reachable_from=self.state))})
        #return attrs
        if (len(self.mutation_permanent)==0):
            final_FP = self.network.fixed(~self.network.obs("exp"))
            ~self.network.obs("init") >= final_FP
            ~self.network.obs("exp") >> "fixpoints" ^ {self.network.obs("exp")};
        else:
            with self.network.mutant(self.mutation_permanent) as m:
                final_FP = m.fixed(~m.obs("exp"))
                ~m.obs("init") >= final_FP
                ~m.obs("exp") >> "fixpoints" ^ {m.obs("exp")};
        if (verbose):
            BNs = list(self.network.boolean_networks(limit=1, njobs=self.njobs))
            nsol = len([bn for bn in tqdm(BNs)])
        else:
            with capture() as out:
                BNs = list(self.network.boolean_networks(limit=1, njobs=self.njobs))
            nsol = len([bn for bn in BNs])
        if (nsol>0):
            attrs = pd.DataFrame({"Attr%d" % self.nattr: self.network.data["exp"]})
        else:
            attrs = pd.DataFrame({"StateNotFound%d" % self.nattr: {"nan":np.nan}})
        self.nattr += 1
        sbcall("rm -f exist_1.zip", shell=True)
        self.attrs = attrs
        return attrs

    def generate_trajectories(self, params={}, outputs=[]):
        raise NotImplemented

############ MABOSS VERSION

class MABOSS_SIM(BN_SIM):
    def __init__(self, seednb=0,njobs=None):
        super(MABOSS_SIM, self).__init__(seednb,njobs)
        self.params = {
            'sample_count': 10000,
            'use_physrandgen': 0,
            'thread_count': self.njobs,
            'max_time': 1000,
            'time_tick': 1,
        }

    def initialize_network(self, network_fname):
        self.network_fname=network_fname
        self.folder=self.network_fname.split("/")[-1].split(".bnet")[0]
        ts = str(int(time()))
        sbcall("cp "+self.network_fname+" network"+ts+".bnet",shell=True)
        sbcall("zip -q -r "+self.folder+".zip network"+ts+".bnet", shell=True)
        self.network = maboss.Ensemble(self.folder+".zip")
        sbcall("rm -f "+self.folder+".zip network"+ts+".bnet", shell=True)
        self.network.param.update(self.params)
        self.network.set_outputs(self.network.nodes)
        self.gene_list = self.network.nodes

    def add_initial_states(self, initial, final=None):
        assert initial.shape[1]==1
        for g in self.network.nodes:
            p_g = 0.5 if (g not in initial.dropna().index) else float(initial.loc[g][initial.columns[0]])
            self.network.set_istate(g, [1.-p_g, p_g]) ## 0: prb(g==0 in initial state), 1: prb(g==1 in initial state)
        for g, p_g in self.mutation_transient.items():
            self.network.set_istate(g, [1.-p_g, p_g])
        self.initial = initial.copy().T
        for g, gval in self.mutation_transient.items():
            self.initial[g] = float(gval)
        self.initial = self.initial.T

    def add_transient_mutation(self, mutation):
        pass

    def add_permanent_mutation(self, mutation):
        self.network.mutations = mutation
        self.network.set_outputs([g for g in self.gene_list if (g not in self.network.mutations)]) 
        mutations_off, mutations_on = [[g for g in mutation if (mutation[g]==v)] for v in [0,1]]
        self.network = maboss.copy_and_mutate(self.network, mutations_off, "OFF")
        self.network = maboss.copy_and_mutate(self.network, mutations_on, "ON")

    def enumerate_attractors(self, verbose=True):
        sbcall("mkdir -p "+self.folder+"/models/",shell=True)
        self.network.write_cfg(self.folder+"/", self.folder+".cfg")
        sbcall("cp "+self.network_fname+" "+self.folder+"/models/"+self.folder+".bnet", shell=True)
        states = self.network.run(workdir=self.folder+"/")._get_raw_states()
        sbcall("rm -rf "+self.folder+"/",shell=True)
        res_mat = np.matrix([[int(g in s) for s in states[-1]] for g in self.gene_list])
        attrs = pd.DataFrame(res_mat, index=self.gene_list, columns=["S%d" % i for i in range(len(states[-1]))]).copy()
        self.attrs = attrs
        return attrs

    def generate_trajectories(self, params={}, outputs=[]):
        self.network.param.update(self.params if (len(params)==0) else params)
        self.network.set_outputs([g for g in (self.gene_list if (len(outputs)==0) else outputs) if (g not in self.network.mutations)])
        sbcall("mkdir -p "+self.folder+"/models/",shell=True)
        self.network.write_cfg(self.folder+"/", self.folder+".cfg")
        sbcall("cp "+self.network_fname+" "+self.folder+"/models/"+self.folder+".bnet", shell=True)
        result = self.network.run(workdir=self.folder+"/")
        result.plot_trajectory()
        table = result.get_last_states_probtraj()
        table.columns = ["{"+",".join([cc+"=1" for cc in list(sorted(c.split(" -- ")))])+"}" if (c!="<nil>") else c for c in table.columns]
        table.index = ["prob"]
        sbcall("rm -rf "+self.folder+"/",shell=True)
        return table



# coding:utf-8

import pandas as pd
import numpy as np

from joblib import Parallel, delayed
import mpbn
import mpbn.simulation as mpbn_sim
from tqdm import tqdm

from sklearn.metrics import pairwise_distances
from sklearn.metrics import accuracy_score as ACC
from sklearn.metrics import roc_auc_score as AUC
from sklearn.metrics import fbeta_score as Fbeta
from random import seed as rseed
from scipy.stats import ks_2samp, spearmanr, kendalltau

#################################################
## Baseline (L1000 CDS^2) [Duan et al., 2016]  ##
#################################################
def baseline(signatures, phenotype, is_binary=False):
    '''
    Compute the cosine scores between a set of signatures S and a differential phenotype P

    ...

    Parameters
    ----------
    S : Pandas DataFrame
        signatures rows/[genes] x columns/[drug names] [Treated || Control]
    P : Pandas DataFrame
        differential phenotype rows/[genes] x column/[disease name] [Diseased || Healthy]
    is_binary : Python bool
        [default=False] : if set to True, the signatures and phenotypes might be binary (that is, with values in {0,1})

    Returns
    -------
    scores : Python float dictionary
        dictionary (keys=drug names, values=scores), the higher the score, the higher the repurposing power
    '''
    SC, PC = signatures.dropna(), phenotype.dropna()
    if (is_binary):
        SC[SC>0] = 1
        PC[PC>0] = 1
        SC[SC<0] = -1
        PC[PC<0] = -1
        SC[pd.isnull(SC)] = 0
        PC[pd.isnull(PC)] = 0
    C = SC.join(PC, how="outer").fillna(0)
    cosscores = 1-pairwise_distances(C[signatures.columns].T,C[phenotype.columns].T, metric="cosine")
    baseline_df = pd.DataFrame(cosscores, columns=["Cosine Score"], index=signatures.columns)
    return baseline_df

#######################################
## Compute accuracy of the predictor ##
#######################################
def empirical_pvalue(sorted_rewards, sorted_ground_truth, nperms, method="ks_2samp"):
    '''
    Compute an empirical p-value corresponding to testing whether the distributions in the predicted scores and the ground truth are similar by randomly permuting the values of the predictions and averaging the number of significant statistics across all permutations

    ...

    Parameters
    ----------
    sorted_rewards : Python float list
        predicted scores sorted by genes of increasing predicted value
    sorted_ground_truth : Python float list
        ground truth scores sorted by genes of increasing predicted value
    nperms : Python integer
        number of permutations to perform
    method : Python character string
        [default="ks_2samp"] : statistical test to perform at each permutation (should belong to scipy.stats)

    Returns
    -------
    pvalue : Python float
        empirical p-value corresponding to the test across all @nperms permutations
    '''
    assert method in ["ks_2samp", "spearmanr", "kendalltau"]
    val_true, _ = eval(method)(sorted_rewards, sorted_ground_truth)
    p_cumsum, N = 0, len(sorted_rewards)
    for nperm in range(nperms):
        rseed(nperm)
        ## Sample a random ordering of elements
        ids__ = np.random.choice(range(N), size=N, replace=False, p=[1/N]*N)
        new_rewards = [sorted_rewards[i] for i in ids__]
        val_new, _ = eval(method)(new_rewards, sorted_ground_truth)
        p_cumsum += int(val_new >= val_true)
    return p_cumsum/nperms

def compute_metrics(rewards, ground_truth, K=[2,5,10], use_negative_class=False, nperms=10000, thres=0., beta=1.):
    '''
    Compute AUC, Hit Ratio @ k of method for positive/negative class with p-value

    ...

    Parameters
    ----------
    rewards : Python float list
        predicted scores
    ground_truth : Python float list
        ground truth scores
    K : Python integer list
        [default=[2,5,10]] : ranks at which the hit ratio should be computed
    use_negative_class : Python bool
        [default=False] : if set to True, compute the performance with respect to the negative class instead of the positive class
    nperms : Python integer
        [default=10000] : number of permutations to perform
    thres : Python float
        [default=0.] : decision threshold to determine the positive (resp. negative) class
    beta : Python float
        [default=1.] : value of the coefficient in the F-measure

    Returns
    -------
    res_di : Python dictionary
        (keys=metrics, values=values of the metrics)
    '''
    ids = np.argsort(rewards).tolist()
    ids.reverse()
    sorted_rewards = [rewards[i] for i in ids]
    sorted_decisions = [int(rewards[i]>thres) for i in ids]
    sorted_ground_truth = [int(ground_truth[i]<0) if (use_negative_class) else int(ground_truth[i]>0) for i in ids]
    hr = [ACC(sorted_ground_truth[:k], sorted_decisions[:k]) for k in K]
    auc = AUC(sorted_ground_truth, sorted_rewards)
    acc = ACC(sorted_ground_truth, sorted_decisions)
    f_score = Fbeta(sorted_ground_truth, sorted_decisions, average='weighted', beta=beta)
    p = empirical_pvalue(sorted_rewards, sorted_ground_truth, nperms)
    res_di = {'AUC': np.round(auc,3), "ACC": np.round(acc,3), "p": np.round(p,3), ("F_%s") % beta: f_score}
    res_di.update({'HR@%d' % k : hr[ik] for ik, k in enumerate(K)})
    return res_di

######################
## DRUG SIMULATOR   ##
######################

def simulate(network_fname, targets, patients, score, simu_params={}, nbseed=0, quiet=False):
    '''
    Simulate and score the individual effects of drugs on patient phenotypes, compared to controls

    ...

    Parameters
    ----------
    network_fname : Python character string
        (relative) path to a network .BNET file
    targets : Pandas DataFrame
        rows/[genes] x columns/[drugs to test] (either 1: active expression, -1: inactive expression, 0: undetermined expression)
    patients : Pandas DataFrame
        rows/[genes] x columns/[samples] (either 1: activatory, -1: inhibitory, 0: no regulation).
    score : Python object
        scoring of attractors
    simu_params : Python dictionary
        [default={}] : arguments to MPBN-SIM
    nbseed : Python integer
        [default=0] : random seed
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    scores : Pandas DataFrame
        rows/[patient phenotypes] x columns/[drug names], values are the associated scores
    '''
    from multiprocessing import cpu_count
    assert simu_params.get('thread_count', 1)>=1 and simu_params.get('thread_count', 1)<=max(1,cpu_count()-2)
    ## Get M30 genes
    with open(network_fname, "r") as f:
        network = str(f.read())
    genes = [x.split(", ")[0] for x in network.split("\n") if (len(x)>0)]
    from random import seed as rseed
    rseed(nbseed)
    np.random.seed(nbseed)
    ## 1. Classification model b/w healthy and patient phenotypes (to classify final attractor states from treated patients)
    ## 2. Compute one score per drug and per patients
    if (simu_params.get('thread_count', 1)==1):
        scores = [simulate_treatment(network_fname, targets.loc[[g for g in targets.index if (g in genes)]], score, patients[[Patient]], simu_params, quiet=quiet) for Patient in patients.columns]
    else:
        scores = Parallel(n_jobs=simu_params['thread_count'], backend='loky')(delayed(simulate_treatment)(network_fname, targets.loc[[g for g in targets.index if (g in genes)]], score, patients[[Patient]], simu_params, quiet=quiet) for Patient in patients.columns)
    scores = pd.DataFrame(scores, index=patients.columns, columns=targets.columns)
    return scores

def compute_frontier(df, samples, nbseed=0, quiet=False):
    '''
    Fit a model to classify control/treated phenotypes

    ...

    Parameters
    ----------
    df : Pandas DataFrame
        rows/[genes] x columns/[samples] (either 1: active expression, -1: inactive expression, 0: undetermined expression)
    samples : Pandas DataFrame
        rows/["annotation"] x columns/[samples ], values are 1 (healthy sample) or 2 (patient sample).
    nbseed : Python integer
        [default=0] : random seed
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    model : Python object
        object with a function "predict" that returns predictions (1: control, or 2: treated) on phenotypes
    '''
    from sklearn import svm
    model = svm.SVC(random_state=nbseed)
    X = df.values.T
    y = samples.values.T
    model.fit(X, y)
    acc = np.mean([int(x==y) for x,y in zip(list(model.predict(X)), list(y.T))])
    if (not quiet):
        print("<NORD_DS> Accuracy of the model %.2f" % acc)
    return model

def compute_score(f, x0, A, score, genes, nb_sims, experiments, repeat=1, exp_name="", quiet=False):
    '''
    Compute similarities between any attractor in WT and in mutants, weighted by their probabilities

    ...

    Parameters
    ----------
    f : Boolean Network (MPBN) object
        the mutated network
    x0 : MPBN object
        initial state
    A : Attractor list
        list of attractors in mutant network
    score : Python object
        scoring of attractors
    genes : Python character string list
        list of genes in the model frontier
    nb_sims : Python integer
        number of iterations to compute the probabilities
    experiments : Python dictionary list
        list of experiments (different rates/depths)
    repeat : Python integer
        [default=1] : how many times should these experiments be repeated
    exp_name : Python character string
        [default=""] : printable for an experiment
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    score : Python float
        change in attractors induced by the mutation
    '''
    for exp in experiments:
        if "name" not in exp:
            continue
        d_scores = []
        for _ in range(repeat):
            rates = getattr(mpbn_sim, f"{exp['rates']}_rates")
            depth = getattr(mpbn_sim, f"{exp['depth']}_depth")
            rates_args = exp.get("rate_args", {})
            depth_args = exp.get("depth_args", {})
            if (not quiet):
                print(exp_name+" "*int(len(exp_name)>0)+(f"- {depth.__name__}{depth_args}\t{rates.__name__}{rates_args}"))
            probs = mpbn_sim.estimate_reachable_attractors_probabilities(f, x0, A, nb_sims, depth(f, **depth_args), rates(f, **rates_args))
            attrs = pd.DataFrame({"MUT_%d"%ia: a for ia, a in enumerate(A)}).replace("*",np.nan).astype(float).loc[genes]
            probs = {i: x for i,x in list(probs.items()) if (x>0)}
            attrs = attrs[[attrs.columns[i] for i in list(probs.keys())]]
            ## if too many attractors, select the most common ones
            if (attrs.shape[1]>45000):
                idx_common = np.argsort([probs[i] for i in range(len(probs))]).tolist()
                idx_common.reverse()
                idx_common = idx_common[:45000]
                attrs = attrs[[attrs.columns[i] for i in idx_common]]
                probs = {i:probs[i] for i in idx_common}
                sprobs = sum(list(probs.values()))
                probs = {i:probs[i]*100/sprobs for i in probs}
            probs = np.array([probs[ia]/100 for ia in probs])
            classification_attrs = score(attrs)
            drug_score = probs.T.dot(classification_attrs)
            d_scores.append(drug_score)
    return np.mean(d_scores) if (repeat>1) else d_scores[0]

def simulate_treatment(network_name, targets, score, state, simu_params={}, quiet=False):
    '''
    Compute the score assigned to a drug with targets in @targets[[drug]] in network

    ...

    Parameters
    ----------
    network_name : Python character string
        filename of the network in .bnet (needs to be pickable)
    targets : Pandas DataFrame
        rows/[genes] x columns/[columns]
    score : Python object
        scoring of attractors
    state : Pandas DataFrame
        binary patient initial state rows/[genes] x columns/[values in {-1,0,1}]
    simu_params : Python dictionary
        [default={}] : arguments to MPBN-SIM
    seednb : Python integer
        [default=0] : random seed
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    effects : Python float dictionary
        distance from attractors from treated networks to control profiles
    '''
    ## 1. Load the Boolean network
    f = mpbn.load(network_name)
    genes = list(state.index)
    ## 2. Create the initial profile
    if (not quiet):
        print("\t<NORD_DS> Initial state %s" % state.columns[0])
    x0 = f.zero()
    for i in list(state.loc[state[state.columns[0]]==1].index):
        x0[i] = 1
    ## 3. Create the mutated networks
    def patch_model(f, patch):
        f = mpbn.MPBooleanNetwork(f)
        for i, fi in patch.items():
            f[i] = fi
        return f
    experiments, nb_sims = [{"name": "mpsim", "rates": simu_params.get("rates", "fully_asynchronous"), "depth": simu_params.get("depth", "constant_unitary")}], simu_params.get("nb_sims", 100)
    mutants = {t: {gx: str(int(targets.loc[gx][t]>0)) for gx in targets[[t]].index if (targets.loc[gx][t]!=0)} for t in targets.columns}
    f_mutants = {name: patch_model(f, patch) for name, patch in mutants.items()}
    ## 4. Get the reachable attractors from initial state in the presence of mutations ("effect during drug exposure")
    ## 5. Estimate probabilities of attractors from Mutants and compute score
    effects = [0 if (t not in f_mutants) else compute_score(f_mutants[t], x0, [a for a in tqdm(list(f_mutants[t].attractors(reachable_from=x0)))], score, genes, nb_sims, experiments, exp_name="Drug %s (%d/%d) in state %s" % (t, it+1, targets.shape[1], state.columns[0]), quiet=quiet) for it, t in enumerate(targets.columns)]
    assert len(effects)==targets.shape[1]
    return effects
